@@ -21,7 +21,7 @@
 //! for larger collections. The API mirrors standard Rust collections.
 
 pub use crate::collections::{
-    hash::{BorrowHashed, Hashed, SmallHashResult},
+    hash::{BorrowHashed, Hashed, StarlarkHashValue},
     hasher::*,
     small_map::{MHIntoIter, MHIter, MHIterMut, SmallMap},
     small_set::SmallSet,
@@ -33,7 +33,6 @@ pub(crate) mod hasher;
 mod idhasher;
 pub mod small_map;
 mod small_set;
-pub(crate) mod stack;
 pub(crate) mod string_pool;
 pub(crate) mod symbol_map;
 pub(crate) mod vec_map;
