@@ -34,10 +34,10 @@ use crate::{
         docs,
         docs::{DocItem, DocString, DocStringKind},
         function::{NativeAttribute, NativeCallableRawDocs},
+        layout::value::ValueLike,
         structs::FrozenStruct,
         types::function::{NativeFunction, NativeMethod},
         AllocFrozenValue, FrozenHeap, FrozenHeapRef, FrozenStringValue, FrozenValue, Heap, Value,
-        ValueLike,
     },
 };
 
@@ -193,6 +193,11 @@ impl Methods {
             .collect()
     }
 
+    pub(crate) fn members(&self) -> impl Iterator<Item = (&str, FrozenValue)> {
+        self.0.members.iter().map(|(k, v)| (k.as_str(), *v))
+    }
+
+    /// Fetch the documentation.
     pub fn documentation(&self) -> DocItem {
         common_documentation(&self.0.docstring, &self.0.members)
     }
@@ -272,7 +277,7 @@ impl GlobalsBuilder {
         match &mut self.struct_fields {
             None => self.variables.insert(name, value),
             Some(fields) => {
-                let name = self.heap.alloc_string_value(name);
+                let name = self.heap.alloc_str(name);
                 fields.insert(name, value)
             }
         };
@@ -289,7 +294,7 @@ impl GlobalsBuilder {
         f: F,
     ) where
         // If I switch this to the trait alias then it fails to resolve the usages
-        F: for<'v> Fn(&mut Evaluator<'v, '_>, Arguments<'v, '_>) -> anyhow::Result<Value<'v>>
+        F: for<'v> Fn(&mut Evaluator<'v, '_>, &Arguments<'v, '_>) -> anyhow::Result<Value<'v>>
             + Send
             + Sync
             + 'static,
@@ -313,6 +318,10 @@ impl GlobalsBuilder {
         value.alloc_frozen_value(&self.heap)
     }
 
+    /// Set per module docstring.
+    ///
+    /// This function is called by the `starlark_derive` generated code
+    /// and rarely needs to be called manually.
     pub fn set_docstring(&mut self, docstring: &str) {
         self.docstring = Some(docstring.to_owned());
     }
@@ -411,7 +420,7 @@ impl MethodsBuilder {
         F: for<'v> Fn(
                 &mut Evaluator<'v, '_>,
                 Value<'v>,
-                Arguments<'v, '_>,
+                &Arguments<'v, '_>,
             ) -> anyhow::Result<Value<'v>>
             + Send
             + Sync
@@ -428,6 +437,11 @@ impl MethodsBuilder {
             }),
         );
     }
+
+    /// Allocate a value using the same underlying heap as the [`MethodsBuilder`]
+    pub fn alloc<'v, V: AllocFrozenValue>(&'v self, value: V) -> FrozenValue {
+        value.alloc_frozen_value(&self.heap)
+    }
 }
 
 /// Used to create methods for a [`StarlarkValue`](crate::values::StarlarkValue).
@@ -437,7 +451,7 @@ impl MethodsBuilder {
 ///
 /// ```ignore
 /// fn my_methods(builder: &mut GlobalsBuilder) {
-///     fn foo(me: ARef<Foo>) -> NoneType {
+///     fn foo(me: ARef<Foo>) -> anyhow::Result<NoneType> {
 ///         ...
 ///     }
 /// }
@@ -549,11 +563,17 @@ fn common_documentation(docstring: &Option<String>, members: &SymbolMap<FrozenVa
 }
 
 #[cfg(test)]
-mod test {
+mod tests {
     use derive_more::Display;
+    use gazebo::any::AnyLifetime;
 
     use super::*;
-    use crate::{assert::Assert, starlark_type, values::StarlarkValue};
+    use crate::{
+        self as starlark,
+        assert::Assert,
+        starlark_type,
+        values::{NoSerialize, StarlarkValue},
+    };
 
     #[test]
     fn test_send_sync()
@@ -564,7 +584,7 @@ mod test {
 
     #[test]
     fn test_set_attribute() {
-        #[derive(Debug, Display)]
+        #[derive(Debug, Display, AnyLifetime, NoSerialize)]
         #[display(fmt = "Magic")]
         struct Magic;
         starlark_simple_value!(Magic);
