@@ -25,20 +25,23 @@ use std::{
 use gazebo::dupe::Dupe;
 
 use crate::{
-    codemap::Span,
     collections::{symbol_map::Symbol, Hashed, SmallMap},
     environment::slots::ModuleSlotId,
     eval::{
         bc::{
             addr::{BcAddr, BcAddrOffset, BcPtrAddr},
-            compiler::call::ArgsCompiledValueBc,
+            call::{BcCallArgsFull, BcCallArgsPos},
             instr::BcInstr,
             instr_impl::InstrDefData,
             opcode::{BcOpcode, BcOpcodeHandler},
+            slow_arg::BcInstrSlowArg,
         },
-        runtime::slots::LocalSlotId,
+        runtime::{call_stack::FrozenFileSpan, slots::LocalSlotId},
+    },
+    values::{
+        known_methods::KnownMethod, typed::FrozenValueTyped, FrozenRef, FrozenStringValue,
+        FrozenValue, StarlarkValue,
     },
-    values::{typed::FrozenValueTyped, FrozenRef, FrozenStringValue, FrozenValue, StarlarkValue},
 };
 
 /// Truncate value if it is too long.
@@ -170,6 +173,40 @@ impl<A: BcInstrArg, B: BcInstrArg, C: BcInstrArg, D: BcInstrArg> BcInstrArg for
     }
 }
 
+#[allow(clippy::many_single_char_names)]
+impl<A: BcInstrArg, B: BcInstrArg, C: BcInstrArg, D: BcInstrArg, E: BcInstrArg> BcInstrArg
+    for (A, B, C, D, E)
+{
+    fn fmt_append((a, b, c, d, e): &Self, ip: BcAddr, f: &mut dyn Write) -> fmt::Result {
+        A::fmt_append(a, ip, f)?;
+        B::fmt_append(b, ip, f)?;
+        C::fmt_append(c, ip, f)?;
+        D::fmt_append(d, ip, f)?;
+        E::fmt_append(e, ip, f)?;
+        Ok(())
+    }
+
+    fn visit_jump_addr((a, b, c, d, e): &Self, consumer: &mut dyn FnMut(BcAddrOffset)) {
+        BcInstrArg::visit_jump_addr(a, consumer);
+        BcInstrArg::visit_jump_addr(b, consumer);
+        BcInstrArg::visit_jump_addr(c, consumer);
+        BcInstrArg::visit_jump_addr(d, consumer);
+        BcInstrArg::visit_jump_addr(e, consumer);
+    }
+
+    fn pops_stack((a, b, c, d, e): &Self) -> u32 {
+        A::pops_stack(a) + B::pops_stack(b) + C::pops_stack(c) + D::pops_stack(d) + E::pops_stack(e)
+    }
+
+    fn pushes_stack((a, b, c, d, e): &Self) -> u32 {
+        A::pushes_stack(a)
+            + B::pushes_stack(b)
+            + C::pushes_stack(c)
+            + D::pushes_stack(d)
+            + E::pushes_stack(e)
+    }
+}
+
 impl<A: BcInstrArg, const N: usize> BcInstrArg for [A; N] {
     fn fmt_append(param: &Self, ip: BcAddr, f: &mut dyn Write) -> fmt::Result {
         for a in param {
@@ -302,9 +339,9 @@ impl BcInstrArg for String {
     }
 }
 
-impl<T: Display> BcInstrArg for FrozenRef<T>
+impl<T: Display> BcInstrArg for FrozenRef<'static, T>
 where
-    FrozenRef<T>: Copy,
+    FrozenRef<'static, T>: Copy,
 {
     fn fmt_append(param: &Self, _ip: BcAddr, f: &mut dyn Write) -> fmt::Result {
         write!(f, " {}", param.as_ref())
@@ -369,9 +406,9 @@ impl BcInstrArg for ModuleSlotId {
     }
 }
 
-impl BcInstrArg for Span {
+impl BcInstrArg for FrozenFileSpan {
     fn fmt_append(param: &Self, _ip: BcAddr, f: &mut dyn Write) -> fmt::Result {
-        write!(f, " {}:{}", param.begin().get(), param.end().get())
+        write!(f, " {}", param)
     }
 
     fn visit_jump_addr(_param: &Self, _consumer: &mut dyn FnMut(BcAddrOffset)) {}
@@ -402,6 +439,22 @@ impl BcInstrArg for BcOpcode {
     }
 }
 
+impl BcInstrArg for KnownMethod {
+    fn fmt_append(_param: &Self, _ip: BcAddr, f: &mut dyn Write) -> fmt::Result {
+        write!(f, " <m>")
+    }
+
+    fn visit_jump_addr(_param: &Self, _consumer: &mut dyn FnMut(BcAddrOffset)) {}
+
+    fn pops_stack(_param: &Self) -> u32 {
+        0
+    }
+
+    fn pushes_stack(_param: &Self) -> u32 {
+        0
+    }
+}
+
 /// Instruction argument encodes how many values are popped
 /// off the stack manually in the instruction impl.
 #[derive(Copy, Clone, Dupe)]
@@ -482,9 +535,9 @@ impl BcInstrArg for ArgPopsStackMaybe1 {
     }
 }
 
-impl BcInstrArg for Vec<(BcAddr, Span)> {
+impl BcInstrArg for Vec<(BcAddr, BcInstrSlowArg)> {
     fn fmt_append(_param: &Self, _ip: BcAddr, f: &mut dyn Write) -> fmt::Result {
-        write!(f, " spans")
+        write!(f, " args")
     }
 
     fn visit_jump_addr(_param: &Self, _consumer: &mut dyn FnMut(BcAddrOffset)) {}
@@ -602,7 +655,7 @@ impl BcInstrArg for InstrDefData {
     }
 }
 
-impl BcInstrArg for ArgsCompiledValueBc {
+impl BcInstrArg for BcCallArgsFull {
     fn fmt_append(param: &Self, _ip: BcAddr, f: &mut dyn Write) -> fmt::Result {
         write!(f, " {{{}}}", param)
     }
@@ -618,6 +671,22 @@ impl BcInstrArg for ArgsCompiledValueBc {
     }
 }
 
+impl BcInstrArg for BcCallArgsPos {
+    fn fmt_append(param: &Self, _ip: BcAddr, f: &mut dyn Write) -> fmt::Result {
+        write!(f, " {}", param.pos)
+    }
+
+    fn visit_jump_addr(_param: &Self, _consumer: &mut dyn FnMut(BcAddrOffset)) {}
+
+    fn pops_stack(param: &Self) -> u32 {
+        param.pos
+    }
+
+    fn pushes_stack(_param: &Self) -> u32 {
+        0
+    }
+}
+
 impl BcOpcode {
     /// Format instruction argument.
     pub(crate) fn fmt_append_arg(
