@@ -30,7 +30,8 @@ use gazebo::{any::AnyLifetime, prelude::*};
 use once_cell::sync::Lazy;
 
 use crate::{
-    codemap::{CodeMap, Span, Spanned},
+    self as starlark,
+    codemap::CodeMap,
     environment::{FrozenModuleRef, Globals},
     eval::{
         bc::{bytecode::Bc, frame::alloca_frame},
@@ -42,9 +43,13 @@ use crate::{
         },
         fragment::{
             expr::ExprCompiled,
+            span::IrSpanned,
             stmt::{OptimizeOnFreezeContext, StmtCompileContext, StmtCompiled, StmtsCompiled},
         },
-        runtime::{arguments::ParametersSpec, evaluator::Evaluator, slots::LocalSlotId},
+        runtime::{
+            arguments::ParametersSpec, call_stack::FrozenFileSpan, evaluator::Evaluator,
+            slots::LocalSlotId,
+        },
         Arguments,
     },
     syntax::ast::ParameterP,
@@ -160,7 +165,7 @@ impl<T> ParameterCompiled<T> {
 #[display(fmt = "DefInfo")]
 pub(crate) struct DefInfo {
     /// Codemap of the file where the function is declared.
-    pub(crate) codemap: CodeMap,
+    pub(crate) codemap: FrozenRef<'static, CodeMap>,
     /// The raw docstring pulled out of the AST.
     pub(crate) docstring: Option<String>,
     pub(crate) scope_names: ScopeNames,
@@ -177,13 +182,14 @@ pub(crate) struct DefInfo {
     pub(crate) inline_def_body: Option<InlineDefBody>,
     /// Globals captured during function or module creation.
     /// Only needed for debugger evaluation.
-    pub(crate) globals: FrozenRef<Globals>,
+    pub(crate) globals: FrozenRef<'static, Globals>,
 }
 
 impl DefInfo {
-    pub(crate) fn empty() -> FrozenRef<DefInfo> {
+    pub(crate) fn empty() -> FrozenRef<'static, DefInfo> {
+        static EMPTY_CODEMAP: Lazy<CodeMap> = Lazy::new(CodeMap::default);
         static EMPTY: Lazy<DefInfo> = Lazy::new(|| DefInfo {
-            codemap: CodeMap::default(),
+            codemap: FrozenRef::new(&EMPTY_CODEMAP),
             docstring: None,
             scope_names: ScopeNames::default(),
             stmt_compiled: Bc::default(),
@@ -196,9 +202,9 @@ impl DefInfo {
     }
 
     pub(crate) fn for_module(
-        codemap: CodeMap,
+        codemap: FrozenRef<'static, CodeMap>,
         scope_names: ScopeNames,
-        globals: FrozenRef<Globals>,
+        globals: FrozenRef<'static, Globals>,
     ) -> DefInfo {
         DefInfo {
             codemap,
@@ -216,9 +222,9 @@ impl DefInfo {
 #[derive(Clone, Debug)]
 pub(crate) struct DefCompiled {
     pub(crate) function_name: String,
-    pub(crate) params: Vec<Spanned<ParameterCompiled<Spanned<ExprCompiled>>>>,
-    pub(crate) return_type: Option<Box<Spanned<ExprCompiled>>>,
-    pub(crate) info: FrozenRef<DefInfo>,
+    pub(crate) params: Vec<IrSpanned<ParameterCompiled<IrSpanned<ExprCompiled>>>>,
+    pub(crate) return_type: Option<Box<IrSpanned<ExprCompiled>>>,
+    pub(crate) info: FrozenRef<'static, DefInfo>,
 }
 
 /// Function body suitable for inlining.
@@ -230,7 +236,7 @@ pub(crate) enum InlineDefBody {
     ///
     /// See the function where this enum variant is computed for the definition
     /// of safe to inline expression.
-    ReturnSafeToInlineExpr(Spanned<ExprCompiled>),
+    ReturnSafeToInlineExpr(IrSpanned<ExprCompiled>),
 }
 
 impl Compiler<'_, '_, '_> {
@@ -243,9 +249,16 @@ impl Compiler<'_, '_, '_> {
         }
     }
 
-    fn parameter(&mut self, x: CstParameter) -> Spanned<ParameterCompiled<Spanned<ExprCompiled>>> {
-        Spanned {
+    fn parameter(
+        &mut self,
+        x: CstParameter,
+    ) -> IrSpanned<ParameterCompiled<IrSpanned<ExprCompiled>>> {
+        let span = FrozenFileSpan {
+            file: self.codemap,
             span: x.span,
+        };
+        IrSpanned {
+            span,
             node: match x.node {
                 ParameterP::Normal(x, t) => {
                     ParameterCompiled::Normal(self.parameter_name(x), self.expr_opt(t))
@@ -269,10 +282,10 @@ impl Compiler<'_, '_, '_> {
     /// If a statement is `return type(x) == "y"` where `x` is a first slot.
     fn is_return_type_is(stmt: &StmtsCompiled) -> Option<FrozenStringValue> {
         match stmt.first().map(|s| &s.node) {
-            Some(StmtCompiled::Return(Spanned {
+            Some(StmtCompiled::Return(IrSpanned {
                 node:
                     ExprCompiled::TypeIs(
-                        box Spanned {
+                        box IrSpanned {
                             // Slot 0 is a slot for the first function argument.
                             node: ExprCompiled::Local(LocalSlotId(0), ..),
                             ..
@@ -345,6 +358,11 @@ impl Compiler<'_, '_, '_> {
                 let y = Compiler::is_safe_to_inline_expr_spanned(y)?;
                 ExprCompiled::Or(box (x, y))
             }
+            ExprCompiled::Seq(box (ref x, ref y)) => {
+                let x = Compiler::is_safe_to_inline_expr_spanned(x)?;
+                let y = Compiler::is_safe_to_inline_expr_spanned(y)?;
+                ExprCompiled::Seq(box (x, y))
+            }
             ExprCompiled::PercentSOne(..) => return None,
             ExprCompiled::FormatOne(box (before, ref v, after)) => {
                 // `FormatOne` is infallible, unlike `PercentSOne`.
@@ -355,25 +373,23 @@ impl Compiler<'_, '_, '_> {
     }
 
     fn is_safe_to_inline_expr_spanned(
-        expr: &Spanned<ExprCompiled>,
-    ) -> Option<Spanned<ExprCompiled>> {
-        Some(Spanned {
+        expr: &IrSpanned<ExprCompiled>,
+    ) -> Option<IrSpanned<ExprCompiled>> {
+        Some(IrSpanned {
             node: Compiler::is_safe_to_inline_expr(&expr.node)?,
-            // We replace span with default because a function can be inlined
-            // into a different file where codemap is different.
-            // Empty span is OK because safe to inline expression is infallible.
-            // In the worst case, it will be a span for the first character in the call site file.
-            span: Span::default(),
+            // Note we are losing the stack trace here.
+            // This is file because inlined expressions are infallible.
+            span: expr.span,
         })
     }
 
     /// Function body is a `return` safe to inline expression (as defined above).
-    fn is_return_safe_to_inline_expr(stmts: &StmtsCompiled) -> Option<Spanned<ExprCompiled>> {
+    fn is_return_safe_to_inline_expr(stmts: &StmtsCompiled) -> Option<IrSpanned<ExprCompiled>> {
         match stmts.first() {
             None => {
                 // Empty function is equivalent to `return None`.
-                Some(Spanned {
-                    span: Span::default(),
+                Some(IrSpanned {
+                    span: FrozenFileSpan::default(),
                     node: ExprCompiled::Value(FrozenValue::new_none()),
                 })
             }
@@ -388,7 +404,7 @@ impl Compiler<'_, '_, '_> {
     }
 
     fn inline_def_body(
-        params: &[Spanned<ParameterCompiled<Spanned<ExprCompiled>>>],
+        params: &[IrSpanned<ParameterCompiled<IrSpanned<ExprCompiled>>>],
         body: &StmtsCompiled,
     ) -> Option<InlineDefBody> {
         if params.len() == 1 && params[0].accepts_positional() {
@@ -427,14 +443,19 @@ impl Compiler<'_, '_, '_> {
         let scope_names = self.exit_scope();
 
         let scope_names = mem::take(scope_names);
+        let local_count = scope_names.used.len().try_into().unwrap();
 
         let inline_def_body = Self::inline_def_body(&params, &body);
 
         let info = self.eval.module_env.frozen_heap().alloc_any(DefInfo {
-            codemap: self.codemap.dupe(),
+            codemap: self.codemap,
             docstring,
             scope_names,
-            stmt_compiled: body.as_bc(&self.compile_context()),
+            stmt_compiled: body.as_bc(
+                &self.compile_context(),
+                local_count,
+                self.eval.module_env.frozen_heap(),
+            ),
             body_stmts: body,
             inline_def_body,
             stmt_compile_context: self.compile_context(),
@@ -452,14 +473,14 @@ impl Compiler<'_, '_, '_> {
 
 /// Starlark function internal representation and implementation of
 /// [`StarlarkValue`].
-#[derive(Derivative)]
+#[derive(Derivative, NoSerialize, AnyLifetime)]
 #[derivative(Debug)]
 pub(crate) struct DefGen<V> {
     parameters: ParametersSpec<V>, // The parameters, **kwargs etc including defaults (which are evaluated afresh each time)
     parameter_captures: Vec<u32>,  // Indices of parameters, which are captured in nested defs
     parameter_types: Vec<(u32, String, V, TypeCompiled)>, // The types of the parameters (sparse indexed array, (0, argm T) implies parameter 0 named arg must have type T)
     return_type: Option<(V, TypeCompiled)>, // The return type annotation for the function
-    pub(crate) def_info: FrozenRef<DefInfo>, // The source code and metadata for this function
+    pub(crate) def_info: FrozenRef<'static, DefInfo>, // The source code and metadata for this function
     /// Any variables captured from the outer scope (nested def/lambda).
     /// Values are either [`Value`] or [`FrozenValu`] pointing respectively to
     /// [`ValueCaptured`] or [`FrozenValueCaptured`].
@@ -492,7 +513,7 @@ impl<'v> Def<'v> {
         parameter_captures: Vec<u32>,
         parameter_types: Vec<(u32, String, Value<'v>, TypeCompiled)>,
         return_type: Option<(Value<'v>, TypeCompiled)>,
-        stmt: FrozenRef<DefInfo>,
+        stmt: FrozenRef<'static, DefInfo>,
         eval: &mut Evaluator<'v, '_>,
     ) -> Value<'v> {
         let captured = stmt
@@ -538,7 +559,6 @@ impl<'v, T1: ValueLike<'v>> DefGen<T1> {
             self.def_info.docstring.as_ref().map(String::as_ref),
         );
 
-
         Some(DocItem::Function(function_docs))
     }
 }
@@ -571,10 +591,10 @@ impl<'v> Freeze for Def<'v> {
         let parameters = self.parameters.freeze(freezer)?;
         let parameter_types = self
             .parameter_types
-            .into_try_map(|(i, s, v, t)| Ok::<_, anyhow::Error>((i, s, v.freeze(freezer)?, t)))?;
+            .into_try_map(|(i, s, v, t)| anyhow::Ok((i, s, v.freeze(freezer)?, t)))?;
         let return_type = self
             .return_type
-            .into_try_map(|(v, t)| Ok::<_, anyhow::Error>((v.freeze(freezer)?, t)))?;
+            .into_try_map(|(v, t)| anyhow::Ok((v.freeze(freezer)?, t)))?;
         let captured = self.captured.try_map(|x| x.freeze(freezer))?;
         let module = AtomicFrozenRefOption::new(self.module.load_relaxed());
         Ok(FrozenDef {
@@ -610,16 +630,15 @@ where
 
     fn invoke(
         &self,
-        me: Value<'v>,
-        location: Option<Span>,
-        args: Arguments<'v, '_>,
+        _me: Value<'v>,
+        args: &Arguments<'v, '_>,
         eval: &mut Evaluator<'v, '_>,
     ) -> anyhow::Result<Value<'v>> {
-        let local_slots = self.def_info.scope_names.used.len() as u32;
-        alloca_frame(eval, local_slots, self.bc().max_stack_size, |eval| {
+        let bc = self.bc();
+        alloca_frame(eval, bc.local_count, bc.max_stack_size, |eval| {
             let slots = eval.current_frame.locals();
             self.parameters.collect_inline(args, slots, eval.heap())?;
-            eval.with_call_stack(me, location, |eval| self.invoke_raw(eval))
+            self.invoke_raw(eval)
         })
     }
 
@@ -664,15 +683,19 @@ where
             eval.wrap_local_slot_captured(LocalSlotId::new(captured));
         }
 
-        // Copy over the parent slots
-        for ((_, me), captured) in self
-            .def_info
-            .scope_names
-            .parent
-            .iter()
-            .zip(self.captured.iter())
-        {
-            eval.current_frame.set_slot(*me, captured.to_value());
+        // Copy over the parent slots.
+        // Explicitly check `self.captured` is not empty to avoid accessing
+        // self.def_info.scope_names which is two indirections.
+        if !self.captured.is_empty() {
+            for ((_, me), captured) in self
+                .def_info
+                .scope_names
+                .parent
+                .iter()
+                .zip(self.captured.iter())
+            {
+                eval.current_frame.set_slot(*me, captured.to_value());
+            }
         }
 
         if Self::FROZEN {
@@ -740,7 +763,11 @@ impl FrozenDef {
                 heap,
                 frozen_heap,
             })
-            .as_bc(&self.def_info.stmt_compile_context);
+            .as_bc(
+                &self.def_info.stmt_compile_context,
+                self.def_info.scope_names.used.len().try_into().unwrap(),
+                frozen_heap,
+            );
 
         // Store the optimized body.
         // This is (relatively) safe because we know that during freeze
