@@ -22,7 +22,7 @@ use gazebo::prelude::*;
 use thiserror::Error;
 
 use crate::{
-    codemap::{Span, Spanned},
+    codemap::Spanned,
     collections::symbol_map::Symbol,
     environment::slots::ModuleSlotId,
     errors::did_you_mean::did_you_mean,
@@ -32,10 +32,11 @@ use crate::{
             Compiler,
         },
         fragment::{
-            call::CallCompiled, compr::ComprCompiled, def::DefCompiled, known::list_to_tuple,
+            call::CallCompiled, compr::ComprCompiled, def::DefCompiled,
+            expr_bool::ExprCompiledBool, known::list_to_tuple, span::IrSpanned,
             stmt::OptimizeOnFreezeContext,
         },
-        runtime::slots::LocalSlotId,
+        runtime::{call_stack::FrozenFileSpan, slots::LocalSlotId},
         FrozenDef,
     },
     syntax::ast::{AstExprP, AstLiteral, AstPayload, AstString, BinOp, ExprP, StmtP},
@@ -137,7 +138,7 @@ impl ExprBinOp {
             ExprBinOp::Divide => a.div(b, heap),
             ExprBinOp::FloorDivide => a.floor_div(b, heap),
             ExprBinOp::BitAnd => a.bit_and(b),
-            ExprBinOp::BitOr => a.bit_or(b),
+            ExprBinOp::BitOr => a.bit_or(b, heap),
             ExprBinOp::BitXor => a.bit_xor(b),
             ExprBinOp::LeftShift => a.left_shift(b),
             ExprBinOp::RightShift => a.right_shift(b),
@@ -154,53 +155,68 @@ pub(crate) enum ExprCompiled {
     LocalCaptured(LocalSlotId),
     Module(ModuleSlotId),
     /// `x == y`
-    Equals(Box<(Spanned<ExprCompiled>, Spanned<ExprCompiled>)>),
+    Equals(Box<(IrSpanned<ExprCompiled>, IrSpanned<ExprCompiled>)>),
     /// `cmp(x <=> y)`
     Compare(
-        Box<(Spanned<ExprCompiled>, Spanned<ExprCompiled>)>,
+        Box<(IrSpanned<ExprCompiled>, IrSpanned<ExprCompiled>)>,
         CompareOp,
     ),
     /// `type(x)`
-    Type(Box<Spanned<ExprCompiled>>),
+    Type(Box<IrSpanned<ExprCompiled>>),
     /// `len(x)`
-    Len(Box<Spanned<ExprCompiled>>),
+    Len(Box<IrSpanned<ExprCompiled>>),
     /// `type(x) == "y"`
-    TypeIs(Box<Spanned<ExprCompiled>>, FrozenStringValue),
-    Tuple(Vec<Spanned<ExprCompiled>>),
-    List(Vec<Spanned<ExprCompiled>>),
-    Dict(Vec<(Spanned<ExprCompiled>, Spanned<ExprCompiled>)>),
+    TypeIs(Box<IrSpanned<ExprCompiled>>, FrozenStringValue),
+    Tuple(Vec<IrSpanned<ExprCompiled>>),
+    List(Vec<IrSpanned<ExprCompiled>>),
+    Dict(Vec<(IrSpanned<ExprCompiled>, IrSpanned<ExprCompiled>)>),
     /// Comprehension.
     Compr(ComprCompiled),
-    Dot(Box<Spanned<ExprCompiled>>, Symbol),
-    ArrayIndirection(Box<(Spanned<ExprCompiled>, Spanned<ExprCompiled>)>),
+    Dot(Box<IrSpanned<ExprCompiled>>, Symbol),
+    ArrayIndirection(Box<(IrSpanned<ExprCompiled>, IrSpanned<ExprCompiled>)>),
     If(
         Box<(
-            Spanned<ExprCompiled>,
-            Spanned<ExprCompiled>,
-            Spanned<ExprCompiled>,
+            IrSpanned<ExprCompiled>,
+            IrSpanned<ExprCompiled>,
+            IrSpanned<ExprCompiled>,
         )>,
     ),
     Slice(
         Box<(
-            Spanned<ExprCompiled>,
-            Option<Spanned<ExprCompiled>>,
-            Option<Spanned<ExprCompiled>>,
-            Option<Spanned<ExprCompiled>>,
+            IrSpanned<ExprCompiled>,
+            Option<IrSpanned<ExprCompiled>>,
+            Option<IrSpanned<ExprCompiled>>,
+            Option<IrSpanned<ExprCompiled>>,
         )>,
     ),
-    Not(Box<Spanned<ExprCompiled>>),
-    UnOp(ExprUnOp, Box<Spanned<ExprCompiled>>),
-    And(Box<(Spanned<ExprCompiled>, Spanned<ExprCompiled>)>),
-    Or(Box<(Spanned<ExprCompiled>, Spanned<ExprCompiled>)>),
+    Not(Box<IrSpanned<ExprCompiled>>),
+    UnOp(ExprUnOp, Box<IrSpanned<ExprCompiled>>),
+    And(Box<(IrSpanned<ExprCompiled>, IrSpanned<ExprCompiled>)>),
+    Or(Box<(IrSpanned<ExprCompiled>, IrSpanned<ExprCompiled>)>),
+    /// Expression equivalent to `(x, y)[1]`: evaluate `x`, discard the result,
+    /// then evaluate `y` and use its result.
+    Seq(Box<(IrSpanned<ExprCompiled>, IrSpanned<ExprCompiled>)>),
     Op(
         ExprBinOp,
-        Box<(Spanned<ExprCompiled>, Spanned<ExprCompiled>)>,
+        Box<(IrSpanned<ExprCompiled>, IrSpanned<ExprCompiled>)>,
     ),
     /// `"aaa%sbbb" % arg`
-    PercentSOne(Box<(FrozenStringValue, Spanned<ExprCompiled>, FrozenStringValue)>),
+    PercentSOne(
+        Box<(
+            FrozenStringValue,
+            IrSpanned<ExprCompiled>,
+            FrozenStringValue,
+        )>,
+    ),
     /// `"aaa%sbbb".format(arg)`
-    FormatOne(Box<(FrozenStringValue, Spanned<ExprCompiled>, FrozenStringValue)>),
-    Call(Spanned<CallCompiled>),
+    FormatOne(
+        Box<(
+            FrozenStringValue,
+            IrSpanned<ExprCompiled>,
+            FrozenStringValue,
+        )>,
+    ),
+    Call(IrSpanned<CallCompiled>),
     Def(DefCompiled),
 }
 
@@ -225,11 +241,6 @@ impl ExprCompiled {
         }
     }
 
-    /// Expression is constant `None`.
-    pub(crate) fn is_none(&self) -> bool {
-        self.as_value().map_or(false, |x| x.is_none())
-    }
-
     /// Is expression a constant string?
     pub(crate) fn as_string(&self) -> Option<FrozenStringValue> {
         FrozenStringValue::new(self.as_value()?)
@@ -282,6 +293,7 @@ impl ExprCompiled {
             Self::Type(x) => x.is_pure_infallible(),
             Self::TypeIs(x, _t) => x.is_pure_infallible(),
             Self::Not(x) => x.is_pure_infallible(),
+            Self::Seq(box (x, y)) => x.is_pure_infallible() && y.is_pure_infallible(),
             Self::Or(box (x, y)) | Self::And(box (x, y)) => {
                 x.is_pure_infallible() && y.is_pure_infallible()
             }
@@ -291,13 +303,50 @@ impl ExprCompiled {
             _ => false,
         }
     }
+
+    /// If this expression is pure, infallible, and known to produce a value,
+    /// return truth of that value.
+    pub(crate) fn is_pure_infallible_to_bool(&self) -> Option<bool> {
+        match self {
+            ExprCompiled::Value(v) => Some(v.to_value().to_bool()),
+            ExprCompiled::List(xs) | ExprCompiled::Tuple(xs)
+                if xs.iter().all(|x| x.is_pure_infallible()) =>
+            {
+                Some(!xs.is_empty())
+            }
+            // TODO(nga): if keys are unique hashable constants, we can fold this to constant too.
+            ExprCompiled::Dict(xs) if xs.is_empty() => Some(false),
+            ExprCompiled::Not(x) => x.is_pure_infallible_to_bool().map(|x| !x),
+            ExprCompiled::And(box (x, y)) => {
+                match (
+                    x.is_pure_infallible_to_bool(),
+                    y.is_pure_infallible_to_bool(),
+                ) {
+                    (Some(true), y) => y,
+                    (Some(false), _) => Some(false),
+                    (None, _) => None,
+                }
+            }
+            ExprCompiled::Or(box (x, y)) => {
+                match (
+                    x.is_pure_infallible_to_bool(),
+                    y.is_pure_infallible_to_bool(),
+                ) {
+                    (Some(false), y) => y,
+                    (Some(true), _) => Some(true),
+                    (None, _) => None,
+                }
+            }
+            _ => None,
+        }
+    }
 }
 
-impl Spanned<ExprCompiled> {
+impl IrSpanned<ExprCompiled> {
     pub(crate) fn optimize_on_freeze(
         &self,
         ctx: &OptimizeOnFreezeContext,
-    ) -> Spanned<ExprCompiled> {
+    ) -> IrSpanned<ExprCompiled> {
         let span = self.span;
         let expr = match self.node {
             ref e @ (ExprCompiled::Value(..)
@@ -377,6 +426,11 @@ impl Spanned<ExprCompiled> {
                 let r = r.optimize_on_freeze(ctx);
                 return ExprCompiled::or(l, r);
             }
+            ExprCompiled::Seq(box (ref l, ref r)) => {
+                let l = l.optimize_on_freeze(ctx);
+                let r = r.optimize_on_freeze(ctx);
+                return ExprCompiled::seq(l, r);
+            }
             ExprCompiled::Op(op, box (ref l, ref r)) => {
                 let l = l.optimize_on_freeze(ctx);
                 let r = r.optimize_on_freeze(ctx);
@@ -393,76 +447,75 @@ impl Spanned<ExprCompiled> {
             ref d @ ExprCompiled::Def(..) => d.clone(),
             ExprCompiled::Call(ref call) => call.optimize_on_freeze(ctx),
         };
-        Spanned { node: expr, span }
+        IrSpanned { node: expr, span }
     }
 }
 
 impl ExprCompiled {
-    fn not(span: Span, expr: Spanned<ExprCompiled>) -> Spanned<ExprCompiled> {
+    fn not(span: FrozenFileSpan, expr: IrSpanned<ExprCompiled>) -> IrSpanned<ExprCompiled> {
         match expr.node {
-            ExprCompiled::Value(x) => Spanned {
+            ExprCompiled::Value(x) => IrSpanned {
                 node: ExprCompiled::Value(FrozenValue::new_bool(!x.to_value().to_bool())),
                 span,
             },
             // Collapse `not not e` to `e` only if `e` is known to produce a boolean.
             ExprCompiled::Not(box ref e) if e.is_definitely_bool() => e.clone(),
-            _ => Spanned {
+            _ => IrSpanned {
                 node: ExprCompiled::Not(box expr),
                 span,
             },
         }
     }
 
-    fn or(l: Spanned<ExprCompiled>, r: Spanned<ExprCompiled>) -> Spanned<ExprCompiled> {
-        let l_span = l.span;
-        if let Some(l) = l.as_value() {
-            if l.to_value().to_bool() {
-                Spanned {
-                    node: ExprCompiled::Value(l),
-                    span: l_span,
-                }
-            } else {
-                r
-            }
+    fn or(l: IrSpanned<ExprCompiled>, r: IrSpanned<ExprCompiled>) -> IrSpanned<ExprCompiled> {
+        if let Some(l_v) = l.is_pure_infallible_to_bool() {
+            if l_v { l } else { r }
         } else {
-            let span = l.span.merge(r.span);
-            Spanned {
+            let span = l.span.merge(&r.span);
+            IrSpanned {
                 node: ExprCompiled::Or(box (l, r)),
                 span,
             }
         }
     }
 
-    fn and(l: Spanned<ExprCompiled>, r: Spanned<ExprCompiled>) -> Spanned<ExprCompiled> {
-        let l_span = l.span;
-        if let Some(l) = l.as_value() {
-            if !l.to_value().to_bool() {
-                Spanned {
-                    node: ExprCompiled::Value(l),
-                    span: l_span,
-                }
-            } else {
-                r
-            }
+    fn and(l: IrSpanned<ExprCompiled>, r: IrSpanned<ExprCompiled>) -> IrSpanned<ExprCompiled> {
+        if let Some(l_v) = l.is_pure_infallible_to_bool() {
+            if l_v { r } else { l }
         } else {
-            let span = l.span.merge(r.span);
-            Spanned {
+            let span = l.span.merge(&r.span);
+            IrSpanned {
                 node: ExprCompiled::And(box (l, r)),
                 span,
             }
         }
     }
 
+    pub(crate) fn seq(
+        l: IrSpanned<ExprCompiled>,
+        r: IrSpanned<ExprCompiled>,
+    ) -> IrSpanned<ExprCompiled> {
+        if l.is_pure_infallible() {
+            r
+        } else {
+            let span = l.span.merge(&r.span);
+            IrSpanned {
+                node: ExprCompiled::Seq(box (l, r)),
+                span,
+            }
+        }
+    }
+
     fn percent(
-        l: Spanned<ExprCompiled>,
-        r: Spanned<ExprCompiled>,
+        l: IrSpanned<ExprCompiled>,
+        r: IrSpanned<ExprCompiled>,
         heap: &Heap,
         frozen_heap: &FrozenHeap,
     ) -> ExprCompiled {
         if let Some(v) = l.as_string() {
             if let Some((before, after)) = parse_percent_s_one(&v) {
-                let before = frozen_heap.alloc_string_value(&before);
-                let after = frozen_heap.alloc_string_value(&after);
+                let before = frozen_heap.alloc_str(&before);
+                let after = frozen_heap.alloc_str(&after);
                 return ExprCompiled::percent_s_one(before, r, after, heap, frozen_heap);
             }
         }
@@ -471,7 +524,7 @@ impl ExprCompiled {
 
     pub(crate) fn percent_s_one(
         before: FrozenStringValue,
-        arg: Spanned<ExprCompiled>,
+        arg: IrSpanned<ExprCompiled>,
         after: FrozenStringValue,
         heap: &Heap,
         frozen_heap: &FrozenHeap,
@@ -479,8 +532,8 @@ impl ExprCompiled {
         if let Some(arg) = arg.as_value() {
             if let Ok(value) = percent_s_one(before.as_str(), arg.to_value(), after.as_str(), heap)
             {
-                let value = frozen_heap.alloc_str(value.unpack_str().unwrap());
-                return ExprCompiled::Value(value);
+                let value = frozen_heap.alloc_str(value.as_str());
+                return ExprCompiled::Value(value.unpack());
             }
         }
 
@@ -489,26 +542,26 @@ impl ExprCompiled {
 
     pub(crate) fn format_one(
         before: FrozenStringValue,
-        arg: Spanned<ExprCompiled>,
+        arg: IrSpanned<ExprCompiled>,
         after: FrozenStringValue,
         heap: &Heap,
         frozen_heap: &FrozenHeap,
     ) -> ExprCompiled {
         if let Some(arg) = arg.as_value() {
             let value = format_one(&before, arg.to_value(), &after, heap);
-            let value = frozen_heap.alloc_str(value.unpack_str().unwrap());
-            return ExprCompiled::Value(value);
+            let value = frozen_heap.alloc_str(value.as_str());
+            return ExprCompiled::Value(value.unpack());
         }
         ExprCompiled::FormatOne(box (before, arg, after))
     }
 
-    fn add(l: Spanned<ExprCompiled>, r: Spanned<ExprCompiled>) -> ExprCompiled {
-        let span = l.span.merge(r.span);
+    fn add(l: IrSpanned<ExprCompiled>, r: IrSpanned<ExprCompiled>) -> ExprCompiled {
+        let span = l.span.merge(&r.span);
         if let (Some(l), Some(r)) = (l.as_list_of_consts(), r.as_list_of_consts()) {
             let lr = l
                 .iter()
                 .chain(r.iter())
-                .map(|x| Spanned {
+                .map(|x| IrSpanned {
                     node: ExprCompiled::Value(*x),
                     span,
                 })
@@ -520,12 +573,12 @@ impl ExprCompiled {
 
     fn bin_op(
         bin_op: ExprBinOp,
-        l: Spanned<ExprCompiled>,
-        r: Spanned<ExprCompiled>,
+        l: IrSpanned<ExprCompiled>,
+        r: IrSpanned<ExprCompiled>,
         heap: &Heap,
         frozen_heap: &FrozenHeap,
     ) -> ExprCompiled {
-        let span = l.span.merge(r.span);
+        let span = l.span.merge(&r.span);
         // Binary operators should have no side effects,
         // but to avoid possible problems, we only fold binary operators on builtin types.
         if let (Some(l), Some(r)) = (l.as_builtin_value(), r.as_builtin_value()) {
@@ -544,38 +597,38 @@ impl ExprCompiled {
     }
 
     fn if_expr(
-        cond: Spanned<ExprCompiled>,
-        t: Spanned<ExprCompiled>,
-        f: Spanned<ExprCompiled>,
-    ) -> Spanned<ExprCompiled> {
-        match cond {
-            Spanned {
-                node: ExprCompiled::Value(cond),
-                ..
-            } => {
-                if cond.to_value().to_bool() {
-                    t
-                } else {
-                    f
+        cond: IrSpanned<ExprCompiled>,
+        t: IrSpanned<ExprCompiled>,
+        f: IrSpanned<ExprCompiled>,
+    ) -> IrSpanned<ExprCompiled> {
+        let cond_span = cond.span;
+        let cond = ExprCompiledBool::new(cond);
+        match cond.node {
+            ExprCompiledBool::Const(true) => t,
+            ExprCompiledBool::Const(false) => f,
+            ExprCompiledBool::Expr(cond) => match cond {
+                ExprCompiled::Not(box cond) => ExprCompiled::if_expr(cond, f, t),
+                ExprCompiled::Seq(box (x, cond)) => {
+                    ExprCompiled::seq(x, ExprCompiled::if_expr(cond, t, f))
                 }
-            }
-            Spanned {
-                node: ExprCompiled::Not(box cond),
-                ..
-            } => Self::if_expr(cond, f, t),
-            cond => {
-                let span = cond.span.merge(t.span).merge(f.span);
-                Spanned {
-                    node: ExprCompiled::If(box (cond, t, f)),
-                    span,
+                cond => {
+                    let cond = IrSpanned {
+                        node: cond,
+                        span: cond_span,
+                    };
+                    let span = cond.span.merge(&t.span).merge(&f.span);
+                    IrSpanned {
+                        node: ExprCompiled::If(box (cond, t, f)),
+                        span,
+                    }
                 }
-            }
+            },
         }
     }
 
     fn un_op(
         op: ExprUnOp,
-        expr: Spanned<ExprCompiled>,
+        expr: IrSpanned<ExprCompiled>,
         heap: &Heap,
         frozen_heap: &FrozenHeap,
     ) -> ExprCompiled {
@@ -590,27 +643,31 @@ impl ExprCompiled {
     }
 
     fn try_values(
-        span: Span,
+        span: FrozenFileSpan,
         values: &[Value],
         heap: &FrozenHeap,
-    ) -> Option<Vec<Spanned<ExprCompiled>>> {
+    ) -> Option<Vec<IrSpanned<ExprCompiled>>> {
         values
             .try_map(|v| {
                 Self::try_value(span, *v, heap)
-                    .map(|expr| Spanned { span, node: expr })
+                    .map(|expr| IrSpanned { span, node: expr })
                     .ok_or(())
             })
             .ok()
     }
 
     /// Try convert a maybe not frozen value to an expression, or discard it.
-    pub(crate) fn try_value(span: Span, v: Value, heap: &FrozenHeap) -> Option<ExprCompiled> {
+    pub(crate) fn try_value(
+        span: FrozenFileSpan,
+        v: Value,
+        heap: &FrozenHeap,
+    ) -> Option<ExprCompiled> {
         if let Some(v) = v.unpack_frozen() {
             // If frozen, we are lucky.
             Some(ExprCompiled::Value(v))
         } else if let Some(v) = v.unpack_str() {
             // If string, copy it to frozen heap.
-            Some(ExprCompiled::Value(heap.alloc_str(v)))
+            Some(ExprCompiled::Value(heap.alloc_str(v).unpack()))
         } else if let Some(v) = v.downcast_ref::<StarlarkFloat>() {
             Some(ExprCompiled::Value(heap.alloc_float(*v)))
         } else if let Some(v) = v.downcast_ref::<Range>() {
@@ -628,8 +685,27 @@ impl ExprCompiled {
         }
     }
 
+    pub(crate) fn compr(compr: ComprCompiled) -> ExprCompiled {
+        match compr {
+            ComprCompiled::List(box x, clauses) => {
+                if clauses.is_nop() {
+                    ExprCompiled::List(Vec::new())
+                } else {
+                    ExprCompiled::Compr(ComprCompiled::List(box x, clauses))
+                }
+            }
+            ComprCompiled::Dict(box (k, v), clauses) => {
+                if clauses.is_nop() {
+                    ExprCompiled::Dict(Vec::new())
+                } else {
+                    ExprCompiled::Compr(ComprCompiled::Dict(box (k, v), clauses))
+                }
+            }
+        }
+    }
+
     /// Construct tuple expression from elements optimizing to frozen tuple value when possible.
-    pub(crate) fn tuple(elems: Vec<Spanned<ExprCompiled>>, heap: &FrozenHeap) -> ExprCompiled {
+    pub(crate) fn tuple(elems: Vec<IrSpanned<ExprCompiled>>, heap: &FrozenHeap) -> ExprCompiled {
         if let Ok(elems) = elems.try_map(|e| e.as_value().ok_or(())) {
             ExprCompiled::Value(heap.alloc_tuple(&elems))
         } else {
@@ -657,7 +733,7 @@ impl ExprCompiled {
     }
 
     fn dot(
-        object: Spanned<ExprCompiled>,
+        object: IrSpanned<ExprCompiled>,
         field: &Symbol,
         heap: &Heap,
         frozen_heap: &FrozenHeap,
@@ -672,11 +748,11 @@ impl ExprCompiled {
     }
 
     fn slice(
-        span: Span,
-        array: Spanned<ExprCompiled>,
-        start: Option<Spanned<ExprCompiled>>,
-        stop: Option<Spanned<ExprCompiled>>,
-        step: Option<Spanned<ExprCompiled>>,
+        span: FrozenFileSpan,
+        array: IrSpanned<ExprCompiled>,
+        start: Option<IrSpanned<ExprCompiled>>,
+        stop: Option<IrSpanned<ExprCompiled>>,
+        step: Option<IrSpanned<ExprCompiled>>,
         heap: &Heap,
         frozen_heap: &FrozenHeap,
     ) -> ExprCompiled {
@@ -701,12 +777,12 @@ impl ExprCompiled {
     }
 
     fn array_indirection(
-        array: Spanned<ExprCompiled>,
-        index: Spanned<ExprCompiled>,
+        array: IrSpanned<ExprCompiled>,
+        index: IrSpanned<ExprCompiled>,
         heap: &Heap,
         frozen_heap: &FrozenHeap,
     ) -> ExprCompiled {
-        let span = array.span.merge(index.span);
+        let span = array.span.merge(&index.span);
         if let (Some(array), Some(index)) = (array.as_builtin_value(), index.as_value()) {
             if let Ok(v) = array.to_value().at(index.to_value(), heap) {
                 if let Some(expr) = ExprCompiled::try_value(span, v, frozen_heap) {
@@ -717,7 +793,7 @@ impl ExprCompiled {
         ExprCompiled::ArrayIndirection(box (array, index))
     }
 
-    pub(crate) fn typ(v: Spanned<ExprCompiled>) -> ExprCompiled {
+    pub(crate) fn typ(v: IrSpanned<ExprCompiled>) -> ExprCompiled {
         match &v.node {
             ExprCompiled::Value(v) => ExprCompiled::Value(v.to_value().get_type_value().unpack()),
             ExprCompiled::Tuple(xs) if xs.iter().all(|e| e.is_pure_infallible()) => {
@@ -742,7 +818,7 @@ impl ExprCompiled {
         }
     }
 
-    pub(crate) fn type_is(v: Spanned<ExprCompiled>, t: FrozenStringValue) -> ExprCompiled {
+    pub(crate) fn type_is(v: IrSpanned<ExprCompiled>, t: FrozenStringValue) -> ExprCompiled {
         if let Some(v) = v.as_value() {
             return ExprCompiled::Value(FrozenValue::new_bool(
                 v.to_value().get_type() == t.as_str(),
@@ -751,7 +827,7 @@ impl ExprCompiled {
         ExprCompiled::TypeIs(box v, t)
     }
 
-    pub(crate) fn len(arg: Spanned<ExprCompiled>) -> ExprCompiled {
+    pub(crate) fn len(arg: IrSpanned<ExprCompiled>) -> ExprCompiled {
         if let Some(arg) = arg.as_value() {
             if let Ok(len) = arg.to_value().length() {
                 return ExprCompiled::Value(FrozenValue::new_int(len));
@@ -760,7 +836,11 @@ impl ExprCompiled {
         ExprCompiled::Len(box arg)
     }
 
-    fn compare(l: Spanned<ExprCompiled>, r: Spanned<ExprCompiled>, cmp: CompareOp) -> ExprCompiled {
+    fn compare(
+        l: IrSpanned<ExprCompiled>,
+        r: IrSpanned<ExprCompiled>,
+        cmp: CompareOp,
+    ) -> ExprCompiled {
         if let (Some(l), Some(r)) = (l.as_value(), r.as_value()) {
             // If comparison fails, let it fail in runtime.
             if let Ok(r) = l.compare(r.to_value()) {
@@ -781,32 +861,32 @@ pub(crate) enum EvalError {
 /// Try fold expression `cmp(l == r)` into `cmp(type(x) == "y")`.
 /// Return original `l` and `r` arguments if fold was unsuccessful.
 fn try_eval_type_is(
-    l: Spanned<ExprCompiled>,
-    r: Spanned<ExprCompiled>,
-) -> Result<Spanned<ExprCompiled>, (Spanned<ExprCompiled>, Spanned<ExprCompiled>)> {
+    l: IrSpanned<ExprCompiled>,
+    r: IrSpanned<ExprCompiled>,
+) -> Result<IrSpanned<ExprCompiled>, (IrSpanned<ExprCompiled>, IrSpanned<ExprCompiled>)> {
     match (l, r) {
         (
-            Spanned {
+            IrSpanned {
                 node: ExprCompiled::Type(l),
                 span: l_span,
             },
-            Spanned {
+            IrSpanned {
                 node: ExprCompiled::Value(r),
                 span: r_span,
             },
         ) => {
             if let Some(r) = FrozenStringValue::new(r) {
-                Ok(Spanned {
+                Ok(IrSpanned {
                     node: ExprCompiled::type_is(*l, r),
-                    span: l_span.merge(r_span),
+                    span: l_span.merge(&r_span),
                 })
             } else {
                 Err((
-                    Spanned {
+                    IrSpanned {
                         node: ExprCompiled::Type(l),
                         span: l_span,
                     },
-                    Spanned {
+                    IrSpanned {
                         node: ExprCompiled::Value(r),
                         span: r_span,
                     },
@@ -817,7 +897,7 @@ fn try_eval_type_is(
     }
 }
 
-fn eval_equals(l: Spanned<ExprCompiled>, r: Spanned<ExprCompiled>) -> ExprCompiled {
+fn eval_equals(l: IrSpanned<ExprCompiled>, r: IrSpanned<ExprCompiled>) -> ExprCompiled {
     if let (Some(l), Some(r)) = (l.as_value(), r.as_value()) {
         // If comparison fails, let it fail in runtime.
         if let Ok(r) = l.equals(r.to_value()) {
@@ -956,7 +1036,7 @@ pub(crate) fn get_attr_hashed_bind<'v>(
 }
 
 impl Compiler<'_, '_, '_> {
-    pub fn expr_opt(&mut self, expr: Option<Box<CstExpr>>) -> Option<Spanned<ExprCompiled>> {
+    pub fn expr_opt(&mut self, expr: Option<Box<CstExpr>>) -> Option<IrSpanned<ExprCompiled>> {
         expr.map(|v| self.expr(*v))
     }
 
@@ -999,9 +1079,12 @@ impl Compiler<'_, '_, '_> {
         }
     }
 
-    pub(crate) fn expr(&mut self, expr: CstExpr) -> Spanned<ExprCompiled> {
+    pub(crate) fn expr(&mut self, expr: CstExpr) -> IrSpanned<ExprCompiled> {
         // println!("compile {}", expr.node);
-        let span = expr.span;
+        let span = FrozenFileSpan {
+            span: expr.span,
+            file: self.codemap,
+        };
         let expr = match expr.node {
             ExprP::Identifier(ident, resolved_ident) => self.expr_ident(ident, resolved_ident),
             ExprP::Lambda(params, box inner, scope_id) => {
@@ -1120,7 +1203,7 @@ impl Compiler<'_, '_, '_> {
                         BinOp::NotEqual => {
                             ExprCompiled::not(
                                 span,
-                                Spanned {
+                                IrSpanned {
                                     span,
                                     node: eval_equals(l, r),
                                 },
@@ -1143,7 +1226,7 @@ impl Compiler<'_, '_, '_> {
                         BinOp::NotIn => {
                             ExprCompiled::not(
                                 span,
-                                Spanned {
+                                IrSpanned {
                                     span,
                                     node: ExprCompiled::bin_op(
                                         ExprBinOp::In,
@@ -1247,6 +1330,13 @@ impl Compiler<'_, '_, '_> {
                 ExprCompiled::Value(val)
             }
         };
-        Spanned { node: expr, span }
+        IrSpanned { node: expr, span }
+    }
+
+    /// Like `expr` but returns an expression optimized assuming
+    /// only the truth of the result is needed.
+    pub(crate) fn expr_truth(&mut self, expr: CstExpr) -> IrSpanned<ExprCompiledBool> {
+        let expr = self.expr(expr);
+        ExprCompiledBool::new(expr)
     }
 }
