@@ -19,7 +19,9 @@ pub(crate) mod call;
 pub(crate) mod compr;
 pub(crate) mod def;
 pub(crate) mod expr;
+pub(crate) mod expr_bool;
 pub(crate) mod known;
 pub(crate) mod module;
 pub(crate) mod small_vec_1;
+pub(crate) mod span;
 pub(crate) mod stmt;
