@@ -26,18 +26,20 @@ use gazebo::{any::AnyLifetime, cast};
 use thiserror::Error;
 
 use crate::{
-    codemap::{FileSpan, Span},
+    codemap::{FileSpan, FileSpanRef},
     collections::{alloca::Alloca, string_pool::StringPool},
     environment::{slots::ModuleSlotId, EnvironmentError, FrozenModuleRef, Module},
     errors::{Diagnostic, Frame},
     eval::{
-        bc::frame::BcFrame,
+        bc::frame::BcFramePtr,
         fragment::def::DefInfo,
         runtime::{
             bc_profile::BcProfile,
-            call_stack::CallStack,
+            before_stmt::BeforeStmt,
+            call_stack::{CallStack, FrozenFileSpan},
             flame_profile::FlameProfile,
             heap_profile::{HeapProfile, HeapProfileFormat},
+            profile::ProfileMode,
             slots::LocalSlotId,
             stmt_profile::StmtProfile,
         },
@@ -48,8 +50,8 @@ use crate::{
         extra::{PrintHandler, StderrPrintHandler},
     },
     values::{
-        recursive_repr_guard::ReprStackReleaseMemoryOnDrop, value_captured_get, FrozenHeap,
-        FrozenRef, Heap, Trace, Tracer, Value, ValueCaptured, ValueLike,
+        value_captured_get, FrozenHeap, FrozenRef, Heap, Trace, Tracer, Value, ValueCaptured,
+        ValueLike,
     },
 };
 
@@ -76,15 +78,13 @@ pub struct Evaluator<'v, 'a> {
     // The module-level variables in scope at the moment.
     // If `None` then we're in the initial module, use variables from `module_env`.
     // If `Some` we've called a `def` in a loaded frozen module.
-    pub(crate) module_variables: Option<FrozenRef<FrozenModuleRef>>,
+    pub(crate) module_variables: Option<FrozenRef<'static, FrozenModuleRef>>,
     /// Current function (`def` or `lambda`) frame: locals and bytecode stack.
-    pub(crate) current_frame: BcFrame<'v>,
+    pub(crate) current_frame: BcFramePtr<'v>,
     // How we deal with a `load` function.
     pub(crate) loader: Option<&'a dyn FileLoader>,
     // `DefInfo` of currently executed function or module.
-    pub(crate) def_info: FrozenRef<DefInfo>,
-    // Make leak sanitizer happy.
-    pub(crate) _repr_stack_release_memory_on_drop: ReprStackReleaseMemoryOnDrop,
+    pub(crate) def_info: FrozenRef<'static, DefInfo>,
     // Should we enable heap profiling or not
     pub(crate) heap_profile: HeapProfile,
     // Should we enable flame profiling or not
@@ -99,7 +99,7 @@ pub struct Evaluator<'v, 'a> {
     // Size of the heap when we should next perform a GC.
     pub(crate) next_gc_level: usize,
     // Extra functions to run on each statement, usually empty
-    pub(crate) before_stmt: Vec<&'a dyn Fn(Span, &mut Evaluator<'v, 'a>)>,
+    pub(crate) before_stmt: BeforeStmt<'v, 'a>,
     // Used for line profiling
     stmt_profile: StmtProfile,
     // Bytecode profile.
@@ -143,20 +143,19 @@ impl<'v, 'a> Evaluator<'v, 'a> {
             call_stack: CallStack::default(),
             module_env: module,
             module_variables: None,
-            current_frame: BcFrame::default(),
+            current_frame: BcFramePtr::null(),
             loader: None,
             extra: None,
             extra_v: None,
             next_gc_level: GC_THRESHOLD,
             disable_gc: false,
             alloca: Alloca::new(),
-            _repr_stack_release_memory_on_drop: ReprStackReleaseMemoryOnDrop,
             heap_profile: HeapProfile::new(),
             stmt_profile: StmtProfile::new(),
             bc_profile: BcProfile::new(),
             flame_profile: FlameProfile::new(),
             heap_or_flame_profile: false,
-            before_stmt: Vec::new(),
+            before_stmt: BeforeStmt::default(),
             def_info: DefInfo::empty(), // Will be replaced before it is used
             string_pool: StringPool::default(),
             breakpoint_handler: None,
@@ -184,109 +183,91 @@ impl<'v, 'a> Evaluator<'v, 'a> {
         self.loader = Some(loader);
     }
 
-    /// Enable profiling, allowing [`Evaluator::write_heap_profile`] to be used.
-    /// Has the side effect of disabling garbage-collection.
+    /// Enable profiling, allowing [`Evaluator::write_profile`] to be used.
+    /// Profilers add overhead, and while some profilers can be used together,
+    /// it's better to run at most one profiler at a time.
+    pub fn enable_profile(&mut self, mode: &ProfileMode) {
+        match mode {
+            ProfileMode::Heap | ProfileMode::HeapFlame => {
+                self.heap_profile.enable();
+                self.heap_or_flame_profile = true;
+                // Disable GC because otherwise why lose the profile records, as we use the heap
+                // to store a complete list of what happened in linear order.
+                self.disable_gc = true;
+            }
+            ProfileMode::Stmt => {
+                self.stmt_profile.enable();
+                self.before_stmt(&|span, eval| eval.stmt_profile.before_stmt(span));
+            }
+            ProfileMode::Flame => {
+                self.flame_profile.enable();
+                self.heap_or_flame_profile = true;
+            }
+            ProfileMode::Bytecode => {
+                self.bc_profile.enable_1();
+            }
+            ProfileMode::BytecodePairs => {
+                self.bc_profile.enable_2();
+            }
+        }
+    }
+
+    /// Generate instructions to invoke before stmt callbacks when evaluating the module,
+    /// even if this module does not use any such callbacks.
     ///
-    /// Starlark contains three types of profile information - `heap`, `stmt` and `flame`.
-    /// These must be enabled _before_ execution with [`enable_heap_profile`](Evaluator::enable_heap_profile)/
-    /// [`enable_stmt_profile`](Evaluator::enable_stmt_profile)/
-    /// [`enable_flame_profile`](Evaluator::enable_flame_profile), then after execution the
-    /// profiles can be written to a file using [`write_heap_profile`](Evaluator::write_heap_profile)/
-    /// [`write_stmt_profile`](Evaluator::write_stmt_profile)/
-    /// [`write_flame_profile`](Evaluator::write_flame_profile). These profiling modes both have
-    /// some overhead, so while they _can_ be used simultaneously, it's usually better to run the
-    /// code repeatedly if that's possible.
+    /// This function need to be called when evaluating a dependency of a module, if a module
+    /// registers `before_stmt` callback.
+    pub fn enable_before_stmt_instrumentation(&mut self) {
+        self.before_stmt.instrument = true;
+    }
+
+    /// Enable instrumentation in module which is loaded by a module to be profiled.
     ///
-    /// * The `heap_profile` mode provides information about the time spent in each function and allocations
-    ///   performed by each function. Enabling this mode the side effect of disabling garbage-collection.
-    ///   This profiling mode is the recommended one.
-    /// * The `stmt_profile` mode provides information about time spent in each statement.
-    /// * The `bc_profile` mode provides information about bytecode instructions.
-    /// * The `flame_profile` and the `heap_profile` mode provide input compatible with
-    ///   [flamegraph.pl](https://github.com/brendangregg/FlameGraph/blob/master/flamegraph.pl).
-    pub fn enable_heap_profile(&mut self) {
-        self.heap_profile.enable();
-        self.heap_or_flame_profile = true;
-        // Disable GC because otherwise why lose the profile records, as we use the heap
-        // to store a complete list of what happened in linear order.
-        self.disable_gc = true;
+    /// This function need to be called when evaluating a dependency of a module, if a module
+    /// does profiling in the given mode.
+    pub fn enable_profile_instrumentation(&mut self, mode: &ProfileMode) {
+        match mode {
+            ProfileMode::Bytecode | ProfileMode::BytecodePairs => {
+                self.bc_profile.enable_1();
+            }
+            _ => {
+                self.before_stmt.instrument = true;
+            }
+        }
     }
 
-    /// Enable statement profiling, allowing [`Evaluator::write_stmt_profile`] to be used.
-    /// See [`Evaluator::enable_heap_profile`] for details about the types of Starlark profiles.
-    pub fn enable_stmt_profile(&mut self) {
-        self.stmt_profile.enable();
-        self.before_stmt(&|span, eval| eval.stmt_profile.before_stmt(span, &eval.def_info.codemap));
-    }
-
-    /// Enable bytecode profiling, allowing [`Evaluator::write_bytecode_profile`] to be used.
-    pub fn enable_bytecode_profile(&mut self) {
-        self.bc_profile.enable_1();
-    }
-
-    /// Enable bytecode instruction pairs profiling,
-    /// allowing [`Evaluator::write_bytecode_profile`] to be used.
-    ///
-    /// This option is mutually exclusive with
-    /// [`enable_bytecode_profile`](Self::enable_bytecode_profile).
-    pub fn enable_bytecode_pairs_profile(&mut self) {
-        self.bc_profile.enable_2();
-    }
-
-    /// Enable statement profiling, allowing [`Evaluator::write_flame_profile`] to be used.
-    /// See [`Evaluator::enable_heap_profile`] for details about the types of Starlark profiles.
-    pub fn enable_flame_profile(&mut self) {
-        self.flame_profile.enable();
-        self.heap_or_flame_profile = true;
-    }
-
-    /// Write a profile (as a summarized `.csv` file) to a file.
-    /// Only valid if [`enable_heap_profile`](Evaluator::enable_heap_profile) was called before execution began.
-    /// See [`Evaluator::enable_heap_profile`] for details about the two types of Starlark profiles.
-    pub fn write_heap_profile<P: AsRef<Path>>(&self, filename: P) -> anyhow::Result<()> {
-        self.heap_profile
-            .write(filename.as_ref(), self.heap(), HeapProfileFormat::Summary)
-            .unwrap_or_else(|| Err(EvaluatorError::HeapProfilingNotEnabled.into()))
-    }
-
-    /// Write a heap profile as a flamegraph, suitable as input to
-    /// [flamegraph.pl](https://github.com/brendangregg/FlameGraph/blob/master/flamegraph.pl).
-    /// Only valid if [`enable_heap_profile`](Evaluator::enable_heap_profile) was called before execution began.
-    /// See [`Evaluator::enable_heap_profile`] for details about the two types of Starlark profiles.
-    pub fn write_heap_flame_profile<P: AsRef<Path>>(&self, filename: P) -> anyhow::Result<()> {
-        self.heap_profile
-            .write(
-                filename.as_ref(),
-                self.heap(),
-                HeapProfileFormat::FlameGraph,
-            )
-            .unwrap_or_else(|| Err(EvaluatorError::HeapProfilingNotEnabled.into()))
-    }
-
-    /// Write a profile (as a `.csv` file) to a file.
-    /// Only valid if [`enable_stmt_profile`](Evaluator::enable_stmt_profile) was called before execution began.
-    /// See [`Evaluator::enable_heap_profile`] for details about two types of Starlark profiles.
-    pub fn write_stmt_profile<P: AsRef<Path>>(&self, filename: P) -> anyhow::Result<()> {
-        self.stmt_profile
-            .write(filename.as_ref())
-            .unwrap_or_else(|| Err(EvaluatorError::StmtProfilingNotEnabled.into()))
-    }
-
-    /// Write a profile (as a `.csv` file) to a file.
-    /// Only valid if [`enable_bc_profile`](Self::enable_bytecode_profile) was called
-    /// before execution began.
-    pub fn write_bytecode_profile<P: AsRef<Path>>(&self, filename: P) -> anyhow::Result<()> {
-        self.bc_profile.write_csv(filename.as_ref())
-    }
-
-    /// Write a profile to a file, suitable as input to
-    /// [flamegraph.pl](https://github.com/brendangregg/FlameGraph/blob/master/flamegraph.pl).
-    /// Only valid if [`enable_flame_profile`](Evaluator::enable_flame_profile) was called before execution began.
-    /// See [`Evaluator::enable_heap_profile`] for details about the types of Starlark profiles.
-    pub fn write_flame_profile<P: AsRef<Path>>(&self, filename: P) -> anyhow::Result<()> {
-        self.flame_profile
-            .write(filename.as_ref())
-            .unwrap_or_else(|| Err(EvaluatorError::FlameProfilingNotEnabled.into()))
+    /// Write a profile to a file.
+    /// Only valid if corresponding profiler was enabled.
+    pub fn write_profile<P: AsRef<Path>>(
+        &self,
+        mode: &ProfileMode,
+        filename: P,
+    ) -> anyhow::Result<()> {
+        match mode {
+            ProfileMode::Heap => self
+                .heap_profile
+                .write(filename.as_ref(), self.heap(), HeapProfileFormat::Summary)
+                .unwrap_or_else(|| Err(EvaluatorError::HeapProfilingNotEnabled.into())),
+            ProfileMode::HeapFlame => self
+                .heap_profile
+                .write(
+                    filename.as_ref(),
+                    self.heap(),
+                    HeapProfileFormat::FlameGraph,
+                )
+                .unwrap_or_else(|| Err(EvaluatorError::HeapProfilingNotEnabled.into())),
+            ProfileMode::Stmt => self
+                .stmt_profile
+                .write(filename.as_ref())
+                .unwrap_or_else(|| Err(EvaluatorError::StmtProfilingNotEnabled.into())),
+            ProfileMode::Bytecode | ProfileMode::BytecodePairs => {
+                self.bc_profile.write_csv(filename.as_ref())
+            }
+            ProfileMode::Flame => self
+                .flame_profile
+                .write(filename.as_ref())
+                .unwrap_or_else(|| Err(EvaluatorError::FlameProfilingNotEnabled.into())),
+        }
     }
 
     /// Enable interactive `breakpoint()`. When enabled, `breakpoint()`
@@ -307,13 +288,13 @@ impl<'v, 'a> Evaluator<'v, 'a> {
         self.call_stack.top_location()
     }
 
-    /// Called before every statement is run with the [`Span`] and a reference to the containing [`Evaluator`].
+    /// Called before every statement is run with the span and a reference to the containing [`Evaluator`].
     /// A list of all possible statements can be obtained in advance by
     /// [`AstModule::stmt_locations`](crate::syntax::AstModule::stmt_locations).
     ///
     /// This function may have no effect is called mid evaluation.
-    pub fn before_stmt(&mut self, f: &'a dyn Fn(Span, &mut Evaluator<'v, 'a>)) {
-        self.before_stmt.push(f)
+    pub fn before_stmt(&mut self, f: &'a dyn Fn(FileSpanRef, &mut Evaluator<'v, 'a>)) {
+        self.before_stmt.before_stmt.push(f)
     }
 
     /// Set the handler invoked when `print` function is used.
@@ -321,13 +302,6 @@ impl<'v, 'a> Evaluator<'v, 'a> {
         self.print_handler = handler;
     }
 
-    /// Given a [`Span`] resolve it to a concrete [`FileSpan`] using
-    /// whatever module is currently at the top of the stack.
-    /// This function can be used in conjunction with [`before_stmt`](Evaluator::before_stmt).
-    pub fn file_span(&self, span: Span) -> FileSpan {
-        self.def_info.codemap.file_span(span)
-    }
-
     pub(crate) fn check_types(&self) -> bool {
         // We currently always check types. We suspect that for performance reasons one day
         // we'll want to make it optional, so guard the relevant places behind this test.
@@ -337,10 +311,10 @@ impl<'v, 'a> Evaluator<'v, 'a> {
     /// Called to add an entry to the call stack, by the function being invoked.
     /// Called for all types of function, including those written in Rust.
     #[inline(always)]
-    pub fn with_call_stack<R>(
+    pub(crate) fn with_call_stack<R>(
         &mut self,
         function: Value<'v>,
-        span: Option<Span>,
+        span: Option<FrozenRef<'static, FrozenFileSpan>>,
         within: impl FnOnce(&mut Self) -> anyhow::Result<R>,
     ) -> anyhow::Result<R> {
         #[cold]
@@ -352,11 +326,7 @@ impl<'v, 'a> Evaluator<'v, 'a> {
             })
         }
 
-        self.call_stack.push(
-            function,
-            span.unwrap_or_default(),
-            span.map(|_| self.def_info),
-        )?;
+        self.call_stack.push(function, span)?;
         if unlikely(self.heap_or_flame_profile) {
             self.heap_profile.record_call_enter(function, self.heap());
             self.flame_profile.record_call_enter(function);
@@ -376,8 +346,8 @@ impl<'v, 'a> Evaluator<'v, 'a> {
     #[inline(always)] // There is only one caller
     pub(crate) fn with_function_context<R>(
         &mut self,
-        module: Option<FrozenRef<FrozenModuleRef>>, // None == use module_env
-        def_info: FrozenRef<DefInfo>,
+        module: Option<FrozenRef<'static, FrozenModuleRef>>, // None == use module_env
+        def_info: FrozenRef<'static, DefInfo>,
         within: impl FnOnce(&mut Self) -> R,
     ) -> R {
         // Capture the variables we will be mutating
