@@ -17,11 +17,13 @@
 
 pub(crate) mod arguments;
 pub(crate) mod bc_profile;
+pub(crate) mod before_stmt;
 pub(crate) mod call_stack;
 pub(crate) mod csv;
 pub(crate) mod evaluator;
 pub(crate) mod file_loader;
 pub(crate) mod flame_profile;
 pub(crate) mod heap_profile;
+pub(crate) mod profile;
 pub(crate) mod slots;
 pub(crate) mod stmt_profile;
