@@ -18,6 +18,7 @@
 use std::cell::Cell;
 
 use crate::{
+    codemap::FileSpanRef,
     environment::{Globals, Module},
     eval::Evaluator,
     syntax::{AstModule, Dialect},
@@ -29,7 +30,7 @@ fn before_stmt() {
     let globals = Globals::new();
     let mut evaluator = Evaluator::new(&module);
     let counter = Cell::new(0);
-    let before_stmt = |_span, _eval: &mut Evaluator<'_, '_>| {
+    let before_stmt = |_span: FileSpanRef, _eval: &mut Evaluator<'_, '_>| {
         counter.set(counter.get() + 1);
     };
     evaluator.before_stmt(&before_stmt);
