@@ -28,7 +28,7 @@ use crate::{
     errors::Diagnostic,
     eval::Evaluator,
     syntax::{AstModule, Dialect},
-    values::{Freeze, Freezer, Heap, SimpleValue, StarlarkValue, Trace, UnpackValue, Value},
+    values::{Freeze, Freezer, Heap, NoSerialize, StarlarkValue, Trace, UnpackValue, Value},
 };
 
 mod basic;
@@ -289,7 +289,7 @@ xs[1] += 1
 fn test_radd() {
     // We want select append to always produce a select, much like the
     // Bazel/Buck `select` function.
-    #[derive(Debug, Display, Clone)]
+    #[derive(Debug, Display, Clone, AnyLifetime, NoSerialize)]
     #[display(fmt = "${:?}", _0)]
     struct Select(Vec<i32>);
     starlark_simple_value!(Select);
@@ -331,7 +331,7 @@ fn test_radd() {
 
     #[starlark_module]
     fn module(build: &mut GlobalsBuilder) {
-        fn select(xs: Vec<i32>) -> Select {
+        fn select(xs: Vec<i32>) -> anyhow::Result<Select> {
             Ok(Select(xs))
         }
     }
@@ -661,7 +661,7 @@ fn test_label_assign() {
     // Test the a.b = c construct.
     // No builtin Starlark types support it, so we have to define a custom type (wapping a dictionary)
 
-    #[derive(Debug, Trace, AnyLifetime, Display)]
+    #[derive(Debug, Trace, AnyLifetime, Display, NoSerialize)]
     #[display(fmt = "{:?}", self)]
     struct Wrapper<'v>(RefCell<SmallMap<String, Value<'v>>>);
 
@@ -681,7 +681,7 @@ fn test_label_assign() {
         }
     }
 
-    #[derive(Debug, AnyLifetime, Display)]
+    #[derive(Debug, AnyLifetime, Display, NoSerialize)]
     #[display(fmt = "FrozenWrapper")]
     struct FrozenWrapper;
 
@@ -689,8 +689,6 @@ fn test_label_assign() {
         starlark_type!("wrapper");
     }
 
-    impl SimpleValue for FrozenWrapper {}
-
     impl<'v> Freeze for Wrapper<'v> {
         type Frozen = FrozenWrapper;
         fn freeze(self, _freezer: &Freezer) -> anyhow::Result<Self::Frozen> {
@@ -700,7 +698,7 @@ fn test_label_assign() {
 
     #[starlark_module]
     fn module(builder: &mut GlobalsBuilder) {
-        fn wrapper() -> Value<'v> {
+        fn wrapper() -> anyhow::Result<Value<'v>> {
             Ok(heap.alloc_complex(Wrapper(RefCell::new(SmallMap::new()))))
         }
     }
@@ -797,6 +795,13 @@ xs[xs] = xs
 #[test]
 fn test_self_mutate_dict() {
     // Check functions that mutate and access self on dicts
+    assert::is_true(
+        r#"
+xs = {1: 2}
+xs |= xs
+xs == {1: 2}
+"#,
+    );
     assert::fail(
         r#"
 xs = {}
@@ -827,6 +832,18 @@ len(xs) == 0
     );
 }
 
+#[test]
+fn test_dict_union() {
+    assert::is_true(
+        r#"
+xs = {1: 2, 3: 4}
+xs |= {5: 6}
+xs |= {1: 7}
+xs.items() == [(1, 7), (3, 4), (5, 6)]
+"#,
+    );
+}
+
 #[test]
 fn test_dict_with_frozen_list_key_inlined() {
     let mut a = Assert::new();
