@@ -29,8 +29,8 @@ use crate::{
     eval::Arguments,
     values::{
         bool::BOOL_TYPE, dict::Dict, float::StarlarkFloat, int::INT_TYPE, list::List,
-        none::NoneType, num::Num, range::Range, string::STRING_TYPE, tuple::Tuple, Heap, Value,
-        ValueError, ValueLike,
+        none::NoneType, num::Num, range::Range, string::STRING_TYPE, tuple::Tuple, Heap,
+        StringValue, Value, ValueError, ValueLike,
     },
 };
 
@@ -67,7 +67,7 @@ pub(crate) fn global_functions(builder: &mut GlobalsBuilder) {
     /// fail("oops", 1, False)  # fail: oops 1 False
     /// # "#, "oops 1 False");
     /// ```
-    fn fail(args: Vec<Value>) -> NoneType {
+    fn fail(args: Vec<Value>) -> anyhow::Result<NoneType> {
         let mut s = String::new();
         for x in args {
             s.push(' ');
@@ -96,7 +96,7 @@ pub(crate) fn global_functions(builder: &mut GlobalsBuilder) {
     /// # "#);
     /// ```
     #[starlark(speculative_exec_safe)]
-    fn any(ref x: Value) -> bool {
+    fn any(ref x: Value) -> anyhow::Result<bool> {
         x.with_iterator(heap, |it| {
             for i in it {
                 if i.to_bool() {
@@ -127,7 +127,7 @@ pub(crate) fn global_functions(builder: &mut GlobalsBuilder) {
     /// # "#);
     /// ```
     #[starlark(speculative_exec_safe)]
-    fn all(ref x: Value) -> bool {
+    fn all(ref x: Value) -> anyhow::Result<bool> {
         x.with_iterator(heap, |it| {
             for i in it {
                 if !i.to_bool() {
@@ -165,7 +165,7 @@ pub(crate) fn global_functions(builder: &mut GlobalsBuilder) {
     /// ```
     #[starlark(type(BOOL_TYPE))]
     #[starlark(speculative_exec_safe)]
-    fn bool(ref x: Option<Value>) -> bool {
+    fn bool(ref x: Option<Value>) -> anyhow::Result<bool> {
         match x {
             None => Ok(false),
             Some(x) => Ok(x.to_bool()),
@@ -190,7 +190,7 @@ pub(crate) fn global_functions(builder: &mut GlobalsBuilder) {
     /// # "#);
     /// ```
     #[starlark(speculative_exec_safe)]
-    fn chr(ref i: Value) -> String {
+    fn chr(ref i: Value) -> anyhow::Result<String> {
         let cp = i.to_int()? as u32;
         match std::char::from_u32(cp) {
             Some(x) => Ok(x.to_string()),
@@ -234,7 +234,7 @@ pub(crate) fn global_functions(builder: &mut GlobalsBuilder) {
     /// ```
     #[starlark(type(Dict::TYPE))]
     #[starlark(speculative_exec_safe)]
-    fn dict(args: Arguments<'v, '_>) -> Dict<'v> {
+    fn dict(args: &Arguments<'v, '_>) -> anyhow::Result<Dict<'v>> {
         // Dict is super hot, and has a slightly odd signature, so we can do a bunch of special cases on it.
         // In particular, we don't generate the kwargs if there are no positional arguments.
         // Therefore we make it take the raw Arguments.
@@ -287,7 +287,7 @@ pub(crate) fn global_functions(builder: &mut GlobalsBuilder) {
     /// # "#);
     /// ```
     #[starlark(speculative_exec_safe)]
-    fn dir(ref x: Value) -> Vec<String> {
+    fn dir(ref x: Value) -> anyhow::Result<Vec<String>> {
         Ok(x.dir_attr())
     }
 
@@ -311,7 +311,7 @@ pub(crate) fn global_functions(builder: &mut GlobalsBuilder) {
     /// # "#);
     /// ```
     #[starlark(speculative_exec_safe)]
-    fn enumerate(ref it: Value, start @ 0: i32) -> Value<'v> {
+    fn enumerate(ref it: Value, start @ 0: i32) -> anyhow::Result<Value<'v>> {
         let v = it
             .iterate(heap)?
             .enumerate()
@@ -348,7 +348,7 @@ pub(crate) fn global_functions(builder: &mut GlobalsBuilder) {
     /// ```
     #[starlark(type(StarlarkFloat::TYPE))]
     #[starlark(speculative_exec_safe)]
-    fn float(ref a: Option<Value>) -> f64 {
+    fn float(ref a: Option<Value>) -> anyhow::Result<f64> {
         if a.is_none() {
             return Ok(0.0);
         }
@@ -392,7 +392,11 @@ pub(crate) fn global_functions(builder: &mut GlobalsBuilder) {
     /// # "#);
     /// ```
     #[starlark(speculative_exec_safe)]
-    fn getattr(ref a: Value, ref attr: &str, ref default: Option<Value>) -> Value<'v> {
+    fn getattr(
+        ref a: Value,
+        ref attr: &str,
+        ref default: Option<Value>,
+    ) -> anyhow::Result<Value<'v>> {
         // Make sure we check if its a function first, to be consistent with `a.f`
         match a.get_attr(attr, heap)? {
             Some(v) => Ok(v),
@@ -409,7 +413,7 @@ pub(crate) fn global_functions(builder: &mut GlobalsBuilder) {
     ///
     /// `hasattr(x, name)` reports whether x has an attribute (field or method)
     /// named `name`.
-    fn hasattr(ref a: Value, ref attr: &str) -> bool {
+    fn hasattr(ref a: Value, ref attr: &str) -> anyhow::Result<bool> {
         Ok(a.has_attr(attr))
     }
 
@@ -429,7 +433,7 @@ pub(crate) fn global_functions(builder: &mut GlobalsBuilder) {
     /// # "#);
     /// ```
     #[starlark(speculative_exec_safe)]
-    fn hash(ref a: &str) -> i32 {
+    fn hash(ref a: &str) -> anyhow::Result<i32> {
         // From the starlark spec:
         // > the hash function for strings is the same as that implemented by java.lang.String.hashCode,
         // > a simple polynomial accumulator over the UTF-16 transcoding of the string:
@@ -508,7 +512,7 @@ pub(crate) fn global_functions(builder: &mut GlobalsBuilder) {
     /// ```
     #[starlark(type(INT_TYPE))]
     #[starlark(speculative_exec_safe)]
-    fn int(ref a: Option<Value>, base: Option<Value>) -> i32 {
+    fn int(ref a: Option<Value>, base: Option<Value>) -> anyhow::Result<i32> {
         if a.is_none() {
             return Ok(0);
         }
@@ -614,7 +618,7 @@ pub(crate) fn global_functions(builder: &mut GlobalsBuilder) {
     /// # "#, "not supported");
     /// ```
     #[starlark(speculative_exec_safe)]
-    fn len(ref a: Value) -> i32 {
+    fn len(ref a: Value) -> anyhow::Result<i32> {
         a.length()
     }
 
@@ -638,7 +642,7 @@ pub(crate) fn global_functions(builder: &mut GlobalsBuilder) {
     /// ```
     #[starlark(type(List::TYPE))]
     #[starlark(speculative_exec_safe)]
-    fn list(ref a: Option<Value>) -> Value<'v> {
+    fn list(ref a: Option<Value>) -> anyhow::Result<Value<'v>> {
         Ok(if let Some(a) = a {
             if let Some(xs) = List::from_value(a) {
                 heap.alloc_list(xs.content())
@@ -670,7 +674,7 @@ pub(crate) fn global_functions(builder: &mut GlobalsBuilder) {
     /// # "#);
     /// ```
     #[starlark(speculative_exec_safe)]
-    fn max(mut args: Vec<Value>, key: Option<Value>) -> Value<'v> {
+    fn max(mut args: Vec<Value>, key: Option<Value>) -> anyhow::Result<Value<'v>> {
         let args = if args.len() == 1 {
             args.swap_remove(0)
         } else {
@@ -694,9 +698,9 @@ pub(crate) fn global_functions(builder: &mut GlobalsBuilder) {
                 }
             }
             Some(key) => {
-                let mut cached = key.invoke_pos(None, &[max], eval)?;
+                let mut cached = key.invoke_pos(&[max], eval)?;
                 for i in it {
-                    let keyi = key.invoke_pos(None, &[i], eval)?;
+                    let keyi = key.invoke_pos(&[i], eval)?;
                     if cached.compare(keyi)? == Ordering::Less {
                         max = i;
                         cached = keyi;
@@ -724,7 +728,7 @@ pub(crate) fn global_functions(builder: &mut GlobalsBuilder) {
     /// # "#);
     /// ```
     #[starlark(speculative_exec_safe)]
-    fn min(mut args: Vec<Value>, key: Option<Value>) -> Value<'v> {
+    fn min(mut args: Vec<Value>, key: Option<Value>) -> anyhow::Result<Value<'v>> {
         let args = if args.len() == 1 {
             args.swap_remove(0)
         } else {
@@ -748,9 +752,9 @@ pub(crate) fn global_functions(builder: &mut GlobalsBuilder) {
                 }
             }
             Some(key) => {
-                let mut cached = key.invoke_pos(None, &[min], eval)?;
+                let mut cached = key.invoke_pos(&[min], eval)?;
                 for i in it {
-                    let keyi = key.invoke_pos(None, &[i], eval)?;
+                    let keyi = key.invoke_pos(&[i], eval)?;
                     if cached.compare(keyi)? == Ordering::Greater {
                         min = i;
                         cached = keyi;
@@ -782,7 +786,7 @@ pub(crate) fn global_functions(builder: &mut GlobalsBuilder) {
     /// # "#);
     /// ```
     #[starlark(speculative_exec_safe)]
-    fn ord(ref a: Value) -> i32 {
+    fn ord(ref a: Value) -> anyhow::Result<i32> {
         if let Some(s) = a.unpack_str() {
             let mut chars = s.chars();
             if let Some(c) = chars.next() {
@@ -831,7 +835,7 @@ pub(crate) fn global_functions(builder: &mut GlobalsBuilder) {
     /// ```
     #[starlark(type(Range::TYPE))]
     #[starlark(speculative_exec_safe)]
-    fn range(ref a1: i32, ref a2: Option<i32>, ref step @ 1: i32) -> Range {
+    fn range(ref a1: i32, ref a2: Option<i32>, ref step @ 1: i32) -> anyhow::Result<Range> {
         let start = match a2 {
             None => 0,
             Some(_) => a1,
@@ -855,12 +859,12 @@ pub(crate) fn global_functions(builder: &mut GlobalsBuilder) {
     /// repr(1)                 == '1'
     /// repr("x")               == "\"x\""
     /// repr([1, "x"])          == "[1, \"x\"]"
-    /// repr("test \"'")        == "\"test \\\"\\'\""
-    /// repr("x\"y😿 \\'")      == "\"x\\\"y😿 \\\\\\'\""
+    /// repr("test \"'")        == "\"test \\\"'\""
+    /// repr("x\"y😿 \\'")      == "\"x\\\"y\\U0001f63f \\\\'\""
     /// "#);
     /// ```
     #[starlark(speculative_exec_safe)]
-    fn repr(ref a: Value) -> Value<'v> {
+    fn repr(ref a: Value) -> anyhow::Result<StringValue<'v>> {
         let mut s = eval.string_pool.alloc();
         a.collect_repr(&mut s);
         let r = eval.heap().alloc_str(&s);
@@ -884,7 +888,7 @@ pub(crate) fn global_functions(builder: &mut GlobalsBuilder) {
     /// # "#);
     /// ```
     #[starlark(speculative_exec_safe)]
-    fn reversed(ref a: Value) -> Value<'v> {
+    fn reversed(ref a: Value) -> anyhow::Result<Value<'v>> {
         let mut v: Vec<Value> = a.iterate(heap)?.collect();
         v.reverse();
         Ok(heap.alloc_list(&v))
@@ -914,14 +918,18 @@ pub(crate) fn global_functions(builder: &mut GlobalsBuilder) {
     /// ```
     // This function is not spec-safe, because it may call `key` function
     // which might be not spec-safe.
-    fn sorted(ref x: Value, key: Option<Value>, reverse: Option<Value>) -> Value<'v> {
+    fn sorted(
+        ref x: Value,
+        key: Option<Value>,
+        reverse: Option<Value>,
+    ) -> anyhow::Result<Value<'v>> {
         let it = x.iterate(heap)?;
         let mut it = match key {
             None => it.map(|x| (x, x)).collect(),
             Some(key) => {
                 let mut v = Vec::new();
                 for el in it {
-                    v.push((el, key.invoke_pos(None, &[el], eval)?));
+                    v.push((el, key.invoke_pos(&[el], eval)?));
                 }
                 v
             }
@@ -967,8 +975,8 @@ pub(crate) fn global_functions(builder: &mut GlobalsBuilder) {
     /// ```
     #[starlark(type(STRING_TYPE))]
     #[starlark(speculative_exec_safe)]
-    fn str(ref a: Value) -> Value<'v> {
-        if a.unpack_str().is_some() {
+    fn str(ref a: Value) -> anyhow::Result<StringValue<'v>> {
+        if let Some(a) = StringValue::new(a) {
             // Special case that can avoid reallocating, but is equivalent.
             Ok(a)
         } else {
@@ -994,7 +1002,7 @@ pub(crate) fn global_functions(builder: &mut GlobalsBuilder) {
     /// ```
     #[starlark(type(Tuple::TYPE))]
     #[starlark(speculative_exec_safe)]
-    fn tuple(ref a: Option<Value>) -> Value<'v> {
+    fn tuple(ref a: Option<Value>) -> anyhow::Result<Value<'v>> {
         let mut l = Vec::new();
         if let Some(a) = a {
             a.with_iterator(heap, |it| {
@@ -1018,7 +1026,7 @@ pub(crate) fn global_functions(builder: &mut GlobalsBuilder) {
     /// # "#);
     /// ```
     #[starlark(speculative_exec_safe)]
-    fn r#type(ref a: Value) -> Value<'v> {
+    fn r#type(ref a: Value) -> anyhow::Result<Value<'v>> {
         Ok(a.get_type_value().unpack().to_value())
     }
 
@@ -1041,7 +1049,7 @@ pub(crate) fn global_functions(builder: &mut GlobalsBuilder) {
     /// # "#);
     /// ```
     #[starlark(speculative_exec_safe)]
-    fn zip(args: Vec<Value>) -> Value<'v> {
+    fn zip(args: Vec<Value>) -> anyhow::Result<Value<'v>> {
         let mut v = Vec::new();
         let mut first = true;
         for arg in args {
