@@ -15,6 +15,11 @@
  * limitations under the License.
  */
 
+//! Types supporting documentation for code written in or for Starlark.
+
+// TODO(nga): document it
+#![allow(missing_docs)]
+
 use std::collections::HashMap;
 
 use gazebo::prelude::*;
@@ -272,7 +277,6 @@ impl DocString {
             let mut current_section = None;
             let mut current_section_text = vec![];
 
-
             for line in details.lines() {
                 if let Some(matches) = section_re.captures(line) {
                     finish_section(&mut current_section, &mut current_section_text);
@@ -314,16 +318,20 @@ impl DocString {
 /// Line / column for where in a file a symbol is.
 #[derive(Debug, Clone, PartialEq, Serialize, Deserialize)]
 pub struct Pos {
+    /// Line number.
+    // TODO(nga): explain is it 0-based or 1-based?
     pub line: usize,
+    /// Column number.
+    // TODO(nga): explain is it 0-based or 1-based?
     pub column: usize,
 }
 
 /// The file a symbol resides in, and if available its location within that file.
-///
-/// `path` should be a string that can be passed into `load()` statements.
 #[derive(Debug, Clone, PartialEq, Serialize, Deserialize)]
 pub struct Location {
+    /// `path` is a string that can be passed into `load()` statements.
     pub path: String,
+    /// Location of the symbol within the file.
     pub position: Option<Pos>,
 }
 
@@ -424,7 +432,6 @@ impl Function {
         }
     }
 
-
     /// Parse out parameter docs from an "Args:" section of a docstring
     ///
     /// `args_section` should be dedented, and generally should just be the `args` key of
@@ -555,16 +562,17 @@ pub struct Doc {
 }
 
 #[cfg(test)]
-mod test {
+mod tests {
     use std::fmt::{Display, Formatter};
 
+    use gazebo::any::AnyLifetime;
     use starlark_derive::starlark_module;
 
     use super::*;
     use crate as starlark;
     use crate::{
         environment::{GlobalsBuilder, GlobalsStatic, Methods, MethodsBuilder, MethodsStatic},
-        values::{StarlarkValue, Value},
+        values::{NoSerialize, StarlarkValue, Value},
     };
 
     /// These are where the module docs go
@@ -575,12 +583,12 @@ mod test {
     fn add_some_value(builder: &mut MethodsBuilder) {
         /// Docs for attr1
         #[starlark(attribute)]
-        fn attr1(_this: Value<'v>) -> String {
+        fn attr1(_this: Value<'v>) -> anyhow::Result<String> {
             Ok("attr1".to_owned())
         }
 
         #[starlark(attribute)]
-        fn attr2(_this: Value<'v>) -> String {
+        fn attr2(_this: Value<'v>) -> anyhow::Result<String> {
             Ok("attr2".to_owned())
         }
 
@@ -591,12 +599,12 @@ mod test {
         ///
         /// # Returns
         /// The string 'func1'
-        fn func1(_this: Value<'v>, foo: String) -> String {
+        fn func1(_this: Value<'v>, foo: String) -> anyhow::Result<String> {
             let _ignore = foo;
             Ok("func1".to_owned())
         }
 
-        fn func2(_this: Value<'v>) -> String {
+        fn func2(_this: Value<'v>) -> anyhow::Result<String> {
             Ok("func2".to_owned())
         }
     }
@@ -614,23 +622,23 @@ mod test {
         ///
         /// # Returns
         /// The string 'func1'
-        fn func1(foo: String) -> String {
+        fn func1(foo: String) -> anyhow::Result<String> {
             let _ignore = foo;
             Ok("func1".to_owned())
         }
 
-        fn func2() -> String {
+        fn func2() -> anyhow::Result<String> {
             Ok("func2".to_owned())
         }
 
         /// A function with only positional arguments.
-        fn func3(ref a1: i32, ref a2: Option<i32>, ref step @ 1: i32) -> String {
+        fn func3(ref a1: i32, ref a2: Option<i32>, ref step @ 1: i32) -> anyhow::Result<String> {
             let _x = (a1, a2, step);
             Ok("func3".to_owned())
         }
     }
 
-    #[derive(Debug)]
+    #[derive(Debug, AnyLifetime, NoSerialize)]
     struct SomeValue {}
 
     impl Display for SomeValue {
@@ -873,7 +881,8 @@ mod test {
             })
         );
         assert_eq!(
-            DocString::from_docstring(DocStringKind::Starlark,
+            DocString::from_docstring(
+                DocStringKind::Starlark,
                 r#"This is a summary line that is not dedented like the 'details'
 
         Typing the first line right after the """ in python docstrings is common,
