@@ -40,7 +40,10 @@ use bumpalo::Bump;
 use either::Either;
 use gazebo::prelude::*;
 
-use crate::values::layout::avalue::{AValue, AValueDyn, BlackHole};
+use crate::values::{
+    layout::avalue::{AValue, AValueDyn, BlackHole},
+    StarlarkValue,
+};
 
 /// Min size of allocated object including header.
 /// Should be able to fit `BlackHole` or forward.
@@ -396,7 +399,22 @@ impl AValueHeader {
         unsafe { &(*self_repr).payload }
     }
 
+    pub(crate) unsafe fn payload<'v, T: StarlarkValue<'v>>(&self) -> &T {
+        debug_assert_eq!(self.unpack().static_type_of_value(), T::static_type_id());
+        &*(self.payload_ptr() as *const T)
+    }
+
     pub(crate) fn unpack<'v>(&'v self) -> &'v dyn AValueDyn<'v> {
+        unsafe {
+            // TODO: this assertion does not belong here.
+            //   Instead, `Value` should be a `Pointer<AValueOrForward>`
+            //   instead of `Pointer<AValueHeader>`,
+            //   and assertion should be where we unpack the pointer.
+            debug_assert!(
+                !(*(self as *const AValueHeader as *const AValueOrForward)).is_forward(),
+                "value is a forward pointer; value cannot be unpacked during GC or freeze"
+            );
+        }
         unsafe {
             let res = &*(from_raw_parts(self.payload_ptr(), self.0));
             mem::transmute::<&'v dyn AValueDyn<'static>, &'v dyn AValueDyn<'v>>(res)
@@ -485,9 +503,7 @@ impl Drop for Arena {
         self.for_each_drop_unordered(|x| {
             // Safe to convert to *mut because we are the only owner
             let x = x.unpack() as *const dyn AValueDyn as *mut dyn AValueDyn;
-            unsafe {
-                ptr::drop_in_place(x)
-            };
+            unsafe { ptr::drop_in_place(x) };
         });
         self.non_drop.reset();
         self.drop.reset();
@@ -495,7 +511,7 @@ impl Drop for Arena {
 }
 
 #[cfg(test)]
-mod test {
+mod tests {
     use super::*;
     use crate::values::{any::StarlarkAny, layout::avalue::simple};
 
