@@ -28,10 +28,10 @@ use std::{
 
 use derive_more::Display;
 use gazebo::{any::AnyLifetime, cast, coerce::Coerce, prelude::*};
+use serde::{Serialize, Serializer};
 
 use crate::{
-    codemap::Span,
-    collections::{SmallHashResult, StarlarkHasher},
+    collections::{StarlarkHashValue, StarlarkHasher},
     environment::Methods,
     eval::{Arguments, Evaluator, FrozenDef},
     values::{
@@ -48,7 +48,7 @@ use crate::{
             array::Array,
             tuple::{FrozenTuple, Tuple},
         },
-        ComplexValue, Freezer, FrozenStringValue, FrozenValue, Heap, SimpleValue, StarlarkValue,
+        ComplexValue, Freezer, FrozenStringValue, FrozenValue, Heap, StarlarkValue,
         StarlarkValueDyn, Trace, Tracer, Value, ValueTyped,
     },
 };
@@ -163,7 +163,7 @@ pub(crate) trait AValue<'v>: StarlarkValueDyn<'v> + Sized {
 
     unsafe fn heap_copy(me: *mut AValueRepr<Self>, tracer: &Tracer<'v>) -> Value<'v>;
 
-    fn get_hash(&self) -> anyhow::Result<SmallHashResult> {
+    fn get_hash(&self) -> anyhow::Result<StarlarkHashValue> {
         let mut hasher = StarlarkHasher::new();
         self.write_hash(&mut hasher)?;
         Ok(hasher.finish_small())
@@ -189,7 +189,7 @@ pub(crate) trait AValueDyn<'v>: StarlarkValueDyn<'v> {
 
     fn is_str(&self) -> bool;
 
-    fn get_hash(&self) -> anyhow::Result<SmallHashResult>;
+    fn get_hash(&self) -> anyhow::Result<StarlarkHashValue>;
 }
 
 impl<'v, A: AValue<'v>> AValueDyn<'v> for A {
@@ -213,7 +213,7 @@ impl<'v, A: AValue<'v>> AValueDyn<'v> for A {
         A::is_str()
     }
 
-    fn get_hash(&self) -> anyhow::Result<SmallHashResult> {
+    fn get_hash(&self) -> anyhow::Result<StarlarkHashValue> {
         self.get_hash()
     }
 }
@@ -273,15 +273,17 @@ pub(crate) fn basic_ref<'v, T: StarlarkValueBasic<'v>>(x: &T) -> &dyn AValueDyn<
     x
 }
 
-pub(crate) fn simple(x: impl SimpleValue) -> impl AValue<'static, ExtraElem = ()> {
+pub(crate) fn simple<T: StarlarkValue<'static>>(x: T) -> impl AValue<'static, ExtraElem = ()> {
+    assert!(!T::is_special());
     AValueImpl(Simple, x)
 }
 
 pub(crate) fn complex<'v, C>(x: C) -> impl AValue<'v, ExtraElem = ()>
 where
     C: ComplexValue<'v>,
-    C::Frozen: SimpleValue,
+    C::Frozen: StarlarkValue<'static>,
 {
+    assert!(!C::is_special());
     AValueImpl(Complex, x)
 }
 
@@ -296,7 +298,7 @@ pub(crate) struct Direct;
 // in the heap (e.g. bool, None)
 pub(crate) struct Basic;
 
-// A type that implements SimpleValue.
+// A non-special type with no references to other Starlark values.
 pub(crate) struct Simple;
 
 // A type that implements ComplexValue.
@@ -344,7 +346,7 @@ impl<'v, T: StarlarkValueBasic<'v>> AValue<'v> for AValueImpl<Basic, T> {
         unreachable!("Basic types don't appear in the heap")
     }
 
-    fn get_hash(&self) -> anyhow::Result<SmallHashResult> {
+    fn get_hash(&self) -> anyhow::Result<StarlarkHashValue> {
         Ok(self.1.get_hash())
     }
 }
@@ -373,8 +375,8 @@ impl<'v> AValue<'v> for AValueImpl<Direct, StarlarkFloat> {
         Self::heap_copy_impl(me, tracer, |_v, _tracer| {})
     }
 
-    fn get_hash(&self) -> anyhow::Result<SmallHashResult> {
-        Ok(Num::from(self.1.0).get_small_hash_result())
+    fn get_hash(&self) -> anyhow::Result<StarlarkHashValue> {
+        Ok(Num::from(self.1.0).get_hash())
     }
 }
 
@@ -426,8 +428,8 @@ impl<'v> AValue<'v> for AValueImpl<Direct, StarlarkStr> {
         v
     }
 
-    fn get_hash(&self) -> anyhow::Result<SmallHashResult> {
-        Ok(self.1.get_small_hash_result())
+    fn get_hash(&self) -> anyhow::Result<StarlarkHashValue> {
+        Ok(self.1.get_hash())
     }
 }
 
@@ -634,8 +636,8 @@ impl<'v> AValue<'v> for AValueImpl<Direct, Array<'v>> {
 }
 
 impl<Mode, C> AValueImpl<Mode, C> {
-    /// `heap_freeze` implementation for `SimpleValue` and `StarlarkFloat`
-    /// (`StarlarkFloat` is logically a simple type, but does not implement `SimpleValue` trait).
+    /// `heap_freeze` implementation for simple `StarlarkValue` and `StarlarkFloat`
+    /// (`StarlarkFloat` is logically a simple type, but it is not considered simple type).
     unsafe fn heap_freeze_simple_impl<'v>(
         me: *mut AValueRepr<Self>,
         freezer: &Freezer,
@@ -650,7 +652,7 @@ impl<Mode, C> AValueImpl<Mode, C> {
     }
 }
 
-impl<'v, T: SimpleValue> AValue<'v> for AValueImpl<Simple, T>
+impl<'v, T: StarlarkValue<'static>> AValue<'v> for AValueImpl<Simple, T>
 where
     'v: 'static,
 {
@@ -700,7 +702,7 @@ impl<Mode, C> AValueImpl<Mode, C> {
 impl<'v, T> AValue<'v> for AValueImpl<Complex, T>
 where
     T: ComplexValue<'v>,
-    T::Frozen: SimpleValue,
+    T::Frozen: StarlarkValue<'static>,
 {
     type StarlarkValue = T;
 
@@ -759,7 +761,7 @@ impl<'v> AValueDyn<'v> for BlackHole {
         false
     }
 
-    fn get_hash(&self) -> anyhow::Result<SmallHashResult> {
+    fn get_hash(&self) -> anyhow::Result<StarlarkHashValue> {
         unreachable!()
     }
 }
@@ -812,9 +814,6 @@ impl<'v> StarlarkValueDyn<'v> for BlackHole {
     fn collect_repr_cycle(&self, _collector: &mut String) {
         panic!()
     }
-    fn collect_json(&self, _collector: &mut String) -> anyhow::Result<()> {
-        panic!()
-    }
     fn to_bool(&self) -> bool {
         panic!()
     }
@@ -836,8 +835,7 @@ impl<'v> StarlarkValueDyn<'v> for BlackHole {
     fn invoke(
         &self,
         _me: Value<'v>,
-        _location: Option<Span>,
-        _args: Arguments<'v, '_>,
+        _args: &Arguments<'v, '_>,
         _eval: &mut Evaluator<'v, '_>,
     ) -> anyhow::Result<Value<'v>> {
         panic!()
@@ -846,8 +844,7 @@ impl<'v> StarlarkValueDyn<'v> for BlackHole {
         &self,
         _me: Value<'v>,
         _this: Value<'v>,
-        _location: Option<Span>,
-        _args: Arguments<'v, '_>,
+        _args: &Arguments<'v, '_>,
         _eval: &mut Evaluator<'v, '_>,
     ) -> anyhow::Result<Value<'v>> {
         panic!()
@@ -925,7 +922,7 @@ impl<'v> StarlarkValueDyn<'v> for BlackHole {
     fn bit_and(&self, _other: Value<'v>) -> anyhow::Result<Value<'v>> {
         panic!()
     }
-    fn bit_or(&self, _other: Value<'v>) -> anyhow::Result<Value<'v>> {
+    fn bit_or(&self, _other: Value<'v>, _heap: &'v Heap) -> anyhow::Result<Value<'v>> {
         panic!()
     }
     fn bit_xor(&self, _other: Value<'v>) -> anyhow::Result<Value<'v>> {
@@ -948,6 +945,15 @@ impl<'v> StarlarkValueDyn<'v> for BlackHole {
     }
 }
 
+impl Serialize for BlackHole {
+    fn serialize<S>(&self, _s: S) -> Result<S::Ok, S::Error>
+    where
+        S: Serializer,
+    {
+        panic!()
+    }
+}
+
 impl<'v, Mode: 'static, T: StarlarkValue<'v>> StarlarkValueDyn<'v> for AValueImpl<Mode, T> {
     fn static_type_id_of_value() -> TypeId
     where
@@ -991,9 +997,6 @@ impl<'v, Mode: 'static, T: StarlarkValue<'v>> StarlarkValueDyn<'v> for AValueImp
     fn collect_repr_cycle(&self, collector: &mut String) {
         self.1.collect_repr_cycle(collector)
     }
-    fn collect_json(&self, collector: &mut String) -> anyhow::Result<()> {
-        self.1.collect_json(collector)
-    }
     fn to_bool(&self) -> bool {
         self.1.to_bool()
     }
@@ -1015,21 +1018,19 @@ impl<'v, Mode: 'static, T: StarlarkValue<'v>> StarlarkValueDyn<'v> for AValueImp
     fn invoke(
         &self,
         me: Value<'v>,
-        location: Option<Span>,
-        args: Arguments<'v, '_>,
+        args: &Arguments<'v, '_>,
         eval: &mut Evaluator<'v, '_>,
     ) -> anyhow::Result<Value<'v>> {
-        self.1.invoke(me, location, args, eval)
+        self.1.invoke(me, args, eval)
     }
     fn invoke_method(
         &self,
         me: Value<'v>,
         this: Value<'v>,
-        location: Option<Span>,
-        args: Arguments<'v, '_>,
+        args: &Arguments<'v, '_>,
         eval: &mut Evaluator<'v, '_>,
     ) -> anyhow::Result<Value<'v>> {
-        self.1.invoke_method(me, this, location, args, eval)
+        self.1.invoke_method(me, this, args, eval)
     }
     fn at(&self, index: Value<'v>, heap: &'v Heap) -> anyhow::Result<Value<'v>> {
         self.1.at(index, heap)
@@ -1104,8 +1105,8 @@ impl<'v, Mode: 'static, T: StarlarkValue<'v>> StarlarkValueDyn<'v> for AValueImp
     fn bit_and(&self, other: Value<'v>) -> anyhow::Result<Value<'v>> {
         self.1.bit_and(other)
     }
-    fn bit_or(&self, other: Value<'v>) -> anyhow::Result<Value<'v>> {
-        self.1.bit_or(other)
+    fn bit_or(&self, other: Value<'v>, heap: &'v Heap) -> anyhow::Result<Value<'v>> {
+        self.1.bit_or(other, heap)
     }
     fn bit_xor(&self, other: Value<'v>) -> anyhow::Result<Value<'v>> {
         self.1.bit_xor(other)
@@ -1127,8 +1128,17 @@ impl<'v, Mode: 'static, T: StarlarkValue<'v>> StarlarkValueDyn<'v> for AValueImp
     }
 }
 
+impl<'v, Mode: 'static, T: StarlarkValue<'v>> Serialize for AValueImpl<Mode, T> {
+    fn serialize<S>(&self, s: S) -> Result<S::Ok, S::Error>
+    where
+        S: Serializer,
+    {
+        erased_serde::serialize(&self.1, s)
+    }
+}
+
 #[cfg(test)]
-mod test {
+mod tests {
     use crate::{environment::Module, values::list::List};
 
     #[test]
