@@ -33,13 +33,14 @@ use std::{
 
 use either::Either;
 use gazebo::{cast, prelude::*};
+use once_cell::sync::Lazy;
 
 use crate::{
-    collections::Hashed,
     eval::FrozenDef,
     values::{
         any::StarlarkAny,
         array::Array,
+        constant_string,
         layout::{
             arena::{AValueHeader, AValueRepr, Arena, HeapSummary, Reservation},
             avalue::{
@@ -47,12 +48,12 @@ use crate::{
                 list_avalue, simple, starlark_str, tuple_avalue, AValue, VALUE_EMPTY_ARRAY,
                 VALUE_EMPTY_FROZEN_LIST, VALUE_EMPTY_TUPLE,
             },
-            constant::constant_string,
+            fast_cell::FastCell,
             value::{FrozenValue, Value},
         },
-        string::hash_string_result,
         types::float::StarlarkFloat,
-        AllocFrozenValue, ComplexValue, FrozenRef, FrozenValueTyped, SimpleValue, ValueTyped,
+        AllocFrozenValue, ComplexValue, FrozenRef, FrozenStringValue, FrozenValueTyped,
+        StarlarkValue, StringValue, StringValueLike, ValueTyped,
     },
 };
 
@@ -61,7 +62,7 @@ use crate::{
 pub struct Heap {
     /// Peak memory seen when a garbage collection takes place (may be lower than currently allocated)
     peak_allocated: Cell<usize>,
-    arena: RefCell<Arena>,
+    arena: FastCell<Arena>,
 }
 
 impl Debug for Heap {
@@ -86,11 +87,16 @@ pub struct FrozenHeap {
 /// `FrozenHeap` when it is no longer modified and can be share between threads.
 /// Although, `arena` is not safe to share between threads, but at least `refs` is.
 #[derive(Default)]
+#[allow(clippy::non_send_fields_in_send_ty)]
 struct FrozenFrozenHeap {
     arena: Arena,
     refs: HashSet<FrozenHeapRef>,
 }
 
+// Safe because we never mutate the Arena other than with &mut
+unsafe impl Sync for FrozenFrozenHeap {}
+unsafe impl Send for FrozenFrozenHeap {}
+
 impl Debug for FrozenHeap {
     fn fmt(&self, f: &mut Formatter<'_>) -> fmt::Result {
         let mut x = f.debug_struct("FrozenHeap");
@@ -109,10 +115,6 @@ impl Debug for FrozenFrozenHeap {
     }
 }
 
-// Safe because we never mutate the Arena other than with &mut
-unsafe impl Sync for FrozenHeapRef {}
-unsafe impl Send for FrozenHeapRef {}
-
 /// A reference to a [`FrozenHeap`] that keeps alive all values on the underlying heap.
 /// Note that the [`Hash`] is consistent for a single [`FrozenHeapRef`], but non-deterministic
 /// across executions and distinct but observably identical [`FrozenHeapRef`] values.
@@ -121,6 +123,20 @@ unsafe impl Send for FrozenHeapRef {}
 // given an underlying FrozenHeap.
 pub struct FrozenHeapRef(Arc<FrozenFrozenHeap>);
 
+fn _test_frozen_heap_ref_send_sync()
+where
+    FrozenHeapRef: Send + Sync,
+{
+}
+
+impl Default for FrozenHeapRef {
+    fn default() -> Self {
+        static EMPTY: Lazy<FrozenHeapRef> =
+            Lazy::new(|| FrozenHeapRef(Arc::new(FrozenFrozenHeap::default())));
+        Lazy::force(&EMPTY).dupe()
+    }
+}
+
 impl Hash for FrozenHeapRef {
     fn hash<H: Hasher>(&self, state: &mut H) {
         let x: &FrozenFrozenHeap = Deref::deref(&self.0);
@@ -188,23 +204,19 @@ impl FrozenHeap {
 
     /// Allocate a string on this heap. Be careful about the warnings
     /// around [`FrozenValue`].
-    pub(crate) fn alloc_str(&self, x: &str) -> FrozenValue {
+    pub fn alloc_str(&self, x: &str) -> FrozenStringValue {
         if let Some(x) = constant_string(x) {
             x
         } else {
             let (v, extra) = self.arena.alloc_extra_non_drop(starlark_str(x.len()));
             MaybeUninit::write_slice(extra, x.as_bytes());
-            FrozenValue::new_repr(unsafe { cast::ptr_lifetime(&*v) })
+            unsafe {
+                FrozenStringValue::new_unchecked(FrozenValue::new_repr(cast::ptr_lifetime(&*v)))
+            }
         }
     }
 
-    /// Allocate a string on this heap and hash it. Be careful about the warnings
-    /// around [`FrozenValue`].
-    pub fn alloc_str_hashed(&self, x: &str) -> Hashed<FrozenValue> {
-        let h = hash_string_result(x);
-        Hashed::new_unchecked(h, self.alloc_str(x))
-    }
-
+    /// Allocate a tuple with the given elements on this heap.
     pub fn alloc_tuple<'v>(&'v self, elems: &[FrozenValue]) -> FrozenValue {
         if elems.is_empty() {
             return FrozenValue::new_repr(&VALUE_EMPTY_TUPLE);
@@ -219,6 +231,7 @@ impl FrozenHeap {
         }
     }
 
+    /// Allocate a list with the given elements on this heap.
     pub fn alloc_list(&self, elems: &[FrozenValue]) -> FrozenValue {
         if elems.is_empty() {
             return FrozenValue::new_repr(&VALUE_EMPTY_FROZEN_LIST);
@@ -237,25 +250,45 @@ impl FrozenHeap {
         self.alloc_raw(float_avalue(f))
     }
 
-    /// Allocate a [`SimpleValue`] on this heap. Be careful about the warnings
-    /// around [`FrozenValue`].
-    pub fn alloc_simple<T: SimpleValue>(&self, val: T) -> FrozenValue {
+    /// Allocate a simple [`StarlarkValue`] on this heap.
+    ///
+    /// Simple value is any starlark value which:
+    /// * bound by `'static` lifetime (in particular, it cannot contain references to other `Value`s)
+    /// * is not special builtin (e.g. `None`)
+    pub fn alloc_simple<T: StarlarkValue<'static>>(&self, val: T) -> FrozenValue {
         self.alloc_raw(simple(val))
     }
 
-    /// Allocate a [`SimpleValue`] and return `FrozenRef` to it.
-    pub(crate) fn alloc_simple_frozen_ref<T: SimpleValue>(&self, value: T) -> FrozenRef<T> {
+    /// Allocate a simple [`StarlarkValue`] and return `FrozenRef` to it.
+    pub(crate) fn alloc_simple_frozen_ref<T: StarlarkValue<'static>>(
+        &self,
+        value: T,
+    ) -> FrozenRef<'static, T> {
         let value = self.alloc_simple(value);
         // Here we could avoid dynamic cast, but this code is not executed frequently.
         value.downcast_frozen_ref().unwrap()
     }
 
     /// Allocate any value in the frozen heap.
-    pub(crate) fn alloc_any<T: Debug + Display + Send + Sync>(&self, value: T) -> FrozenRef<T> {
+    pub(crate) fn alloc_any<T: Debug + Display + Send + Sync>(
+        &self,
+        value: T,
+    ) -> FrozenRef<'static, T> {
         let value = self.alloc_simple_frozen_ref(StarlarkAny::new(value));
         value.map(|r| &r.0)
     }
 
+    /// Allocate any value, use `Debug` implementation for `Display`.
+    pub(crate) fn alloc_any_display_from_debug<T: Debug + Send + Sync>(
+        &self,
+        value: T,
+    ) -> FrozenRef<'static, T> {
+        #[derive(derive_more::Display, Debug)]
+        #[display(fmt = "{:?}", _0)]
+        struct Wrapper<T: Debug + Send + Sync>(T);
+        self.alloc_any(Wrapper(value)).map(|r| &r.0)
+    }
+
     /// Number of bytes allocated on this heap, not including any memory
     /// represented by [`extra_memory`](crate::values::StarlarkValue::extra_memory).
     pub fn allocated_bytes(&self) -> usize {
@@ -281,7 +314,7 @@ pub struct Freezer {
     /// Freezing into this heap.
     pub(crate) heap: FrozenHeap,
     /// Defs frozen by this freezer.
-    pub(crate) frozen_defs: RefCell<Vec<FrozenRef<FrozenDef>>>,
+    pub(crate) frozen_defs: RefCell<Vec<FrozenRef<'static, FrozenDef>>>,
 }
 
 impl Freezer {
@@ -364,7 +397,7 @@ impl Heap {
     }
 
     fn alloc_raw<'v, 'v2: 'v2>(&'v self, x: impl AValue<'v2, ExtraElem = ()>) -> Value<'v> {
-        let arena_ref = self.arena.borrow_mut();
+        let arena_ref = self.arena.borrow();
         let arena = &*arena_ref;
         let v: &AValueRepr<_> = arena.alloc(x);
 
@@ -388,8 +421,8 @@ impl Heap {
         &'v self,
         len: usize,
         init: impl FnOnce(*mut u8),
-    ) -> Value<'v> {
-        let arena_ref = self.arena.borrow_mut();
+    ) -> StringValue<'v> {
+        let arena_ref = self.arena.borrow();
         let arena = &*arena_ref;
         let (v, extra) = arena.alloc_extra_non_drop::<_>(starlark_str(len));
         init(extra.as_mut_ptr() as *mut u8);
@@ -397,13 +430,16 @@ impl Heap {
         // We have an arena inside a RefCell which stores ValueMem<'v>
         // However, we promise not to clear the RefCell other than for GC
         // so we can make the `arena` available longer
-        unsafe { transmute!(Value, Value, Value::new_repr(&*v)) }
+        unsafe {
+            let value = transmute!(Value, Value, Value::new_repr(&*v));
+            StringValue::new_unchecked(value)
+        }
     }
 
     /// Allocate a string on the heap.
-    pub fn alloc_str<'v>(&'v self, x: &str) -> Value<'v> {
+    pub fn alloc_str<'v>(&'v self, x: &str) -> StringValue<'v> {
         if let Some(x) = constant_string(x) {
-            x.to_value()
+            x.to_string_value()
         } else {
             self.alloc_str_init(x.len(), |dest| unsafe {
                 copy_nonoverlapping(x.as_ptr(), dest, x.len())
@@ -411,13 +447,8 @@ impl Heap {
         }
     }
 
-    /// Allocate a string on this heap and hash it.
-    pub fn alloc_str_hashed<'v>(&'v self, x: &str) -> Hashed<Value<'v>> {
-        let h = hash_string_result(x);
-        Hashed::new_unchecked(h, self.alloc_str(x))
-    }
-
-    pub(crate) fn alloc_str_concat<'v>(&'v self, x: &str, y: &str) -> Value<'v> {
+    /// Allocate a string on the heap, based on two concatenated strings.
+    pub fn alloc_str_concat<'v>(&'v self, x: &str, y: &str) -> StringValue<'v> {
         if x.is_empty() {
             self.alloc_str(y)
         } else if y.is_empty() {
@@ -430,7 +461,8 @@ impl Heap {
         }
     }
 
-    pub(crate) fn alloc_str_concat3<'v>(&'v self, x: &str, y: &str, z: &str) -> Value<'v> {
+    /// Allocate a string on the heap, based on three concatenated strings.
+    pub fn alloc_str_concat3<'v>(&'v self, x: &str, y: &str, z: &str) -> StringValue<'v> {
         if x.is_empty() {
             self.alloc_str_concat(y, z)
         } else if y.is_empty() {
@@ -448,6 +480,7 @@ impl Heap {
         }
     }
 
+    /// Allocate a tuple with the given elements.
     pub fn alloc_tuple<'v>(&'v self, elems: &[Value<'v>]) -> Value<'v> {
         if elems.is_empty() {
             return FrozenValue::new_repr(&VALUE_EMPTY_TUPLE).to_value();
@@ -475,12 +508,14 @@ impl Heap {
         }
     }
 
+    /// Allocate a list with the given elements.
     pub fn alloc_list<'v>(&'v self, elems: &[Value<'v>]) -> Value<'v> {
         let array = self.alloc_array(elems.len());
         array.extend_from_slice(elems);
         self.alloc_raw(list_avalue(array))
     }
 
+    /// Allocate a list with the given elements.
     pub fn alloc_list_iter<'v>(&'v self, elems: impl IntoIterator<Item = Value<'v>>) -> Value<'v> {
         let elems = elems.into_iter();
         let array = self.alloc_array(0);
@@ -497,7 +532,7 @@ impl Heap {
         self.alloc_raw(list_avalue(array))
     }
 
-    pub(crate) fn alloc_char<'v>(&'v self, x: char) -> Value<'v> {
+    pub(crate) fn alloc_char<'v>(&'v self, x: char) -> StringValue<'v> {
         let mut dst = [0; 4];
         let res = x.encode_utf8(&mut dst);
         self.alloc_str(res)
@@ -507,8 +542,12 @@ impl Heap {
         self.alloc_raw(float_avalue(f))
     }
 
-    /// Allocate a [`SimpleValue`] on the [`Heap`].
-    pub fn alloc_simple<'v, T: SimpleValue>(&'v self, x: T) -> Value<'v> {
+    /// Allocate a simple [`StarlarkValue`] on this heap.
+    ///
+    /// Simple value is any starlark value which:
+    /// * bound by `'static` lifetime (in particular, it cannot contain references to other `Value`s)
+    /// * is not special builtin (e.g. `None`)
+    pub fn alloc_simple<'v, T: StarlarkValue<'static>>(&'v self, x: T) -> Value<'v> {
         self.alloc_raw(simple(x))
     }
 
@@ -516,13 +555,13 @@ impl Heap {
     pub fn alloc_complex<'v, T>(&'v self, x: T) -> Value<'v>
     where
         T: ComplexValue<'v>,
-        T::Frozen: SimpleValue,
+        T::Frozen: StarlarkValue<'static>,
     {
         self.alloc_raw(complex(x))
     }
 
-    pub(crate) fn for_each_ordered<'v>(&'v self, mut f: impl FnMut(Value<'v>)) {
-        self.arena.borrow_mut().for_each_ordered(|x| {
+    pub(crate) unsafe fn for_each_ordered<'v>(&'v self, mut f: impl FnMut(Value<'v>)) {
+        (*self.arena.get_mut()).for_each_ordered(|x| {
             // Otherwise the Value is constrainted by the borrow_mut, when
             // we consider values to be kept alive permanently, other than
             // when a GC happens
@@ -542,16 +581,18 @@ impl Heap {
         self.garbage_collect_internal(f)
     }
 
-    fn garbage_collect_internal<'v>(&'v self, f: impl FnOnce(&Tracer<'v>)) {
-        // Must rewrite all Value's so they point at the new heap
-        let mut arena = self.arena.borrow_mut();
+    unsafe fn garbage_collect_internal<'v>(&'v self, f: impl FnOnce(&Tracer<'v>)) {
+        // Must rewrite all Value's so they point at the new heap.
+        // Take the arena out of the heap to make sure nobody allocates in it,
+        // but hold the reference until the GC is done.
+        let _arena = self.arena.take();
 
         let tracer = Tracer::<'v> {
             arena: Arena::default(),
             phantom: PhantomData,
         };
         f(&tracer);
-        *arena = tracer.arena;
+        self.arena.set(tracer.arena);
     }
 
     /// Obtain a summary of how much memory is currently allocated by this heap.
