@@ -18,19 +18,20 @@
 // Possible optimisations:
 // Encoding none, bool etc in the pointer of frozen value
 
-pub(crate) use constant::StringValueLike;
-pub use constant::{FrozenStringValue, StarlarkStrNRepr, StringValue};
 pub use heap::{Freezer, FrozenHeap, FrozenHeapRef, Heap, Tracer};
 pub(crate) use pointer_i32::PointerI32;
-pub use value::{FrozenValue, Value, ValueIdentity};
+pub(crate) use string::StringValueLike;
+pub use string::{static_string::*, FrozenStringValue, StringValue};
 pub(crate) use value_captured::*;
 
 mod arena;
 mod avalue;
-mod constant;
+mod fast_cell;
 mod heap;
+pub(crate) mod identity;
 mod pointer;
 mod pointer_i32;
+mod string;
 pub(crate) mod typed;
-mod value;
+pub(crate) mod value;
 mod value_captured;
