@@ -35,10 +35,11 @@ use std::{
 };
 
 use derive_more::Display;
+use erased_serde::Serialize;
 use gazebo::any::AnyLifetime;
 
 use crate::{
-    codemap::Span,
+    self as starlark,
     collections::StarlarkHasher,
     environment::Methods,
     eval::{Arguments, Evaluator},
@@ -63,11 +64,11 @@ use crate::{
 /// generate `One` and `FrozenOne` aliases.
 ///
 /// ```
-/// use starlark::values::{AnyLifetime, ComplexValue, Coerce, Freezer, FrozenValue, SimpleValue, StarlarkValue, Value, ValueLike, Trace, Tracer, Freeze};
+/// use starlark::values::{AnyLifetime, ComplexValue, Coerce, Freezer, FrozenValue, StarlarkValue, Value, ValueLike, Trace, Tracer, Freeze, NoSerialize};
 /// use starlark::{starlark_complex_value, starlark_type};
 /// use derive_more::Display;
 ///
-/// #[derive(Debug, Trace, Coerce, Display)]
+/// #[derive(Debug, Trace, Coerce, Display, AnyLifetime, NoSerialize)]
 /// #[repr(C)]
 /// struct OneGen<V>(V);
 /// starlark_complex_value!(One);
@@ -105,8 +106,7 @@ use crate::{
 ///
 /// The macro also defines instances of [`AnyLifetime`] for both,
 /// [`AllocValue`](crate::values::AllocValue) for both,
-/// [`AllocFrozenValue`](crate::values::AllocFrozenValue) for the frozen one,
-/// [`SimpleValue`] for the frozen one and
+/// [`AllocFrozenValue`](crate::values::AllocFrozenValue) for the frozen one, and
 /// [`UnpackValue`](crate::values::UnpackValue) for the non-frozen one.
 /// It also defines the methods:
 ///
@@ -118,7 +118,7 @@ use crate::{
 /// impl<'v> One<'v> {
 ///     // Obtain a reference to `One` from a `Value`, regardless
 ///     // of whether the underlying `Value` is a `One` or `FrozenOne`.
-///     pub fn from_value(x: Value<'v>) -> Option<ARef<'v, Self>> {
+///     pub fn from_value(x: Value<'v>) -> Option<&'v Self> {
 /// # unimplemented!(
 /// # r#"
 ///         ...
@@ -146,86 +146,36 @@ use crate::{
 ///   when non-frozen and a direct value when frozen.
 pub trait ComplexValue<'v>: StarlarkValue<'v> + Trace<'v> + Freeze
 where
-    <Self as Freeze>::Frozen: SimpleValue,
+    <Self as Freeze>::Frozen: StarlarkValue<'static>,
 {
 }
 
 impl<'v, V> ComplexValue<'v> for V
 where
     V: StarlarkValue<'v> + Trace<'v> + Freeze,
-    <V as Freeze>::Frozen: SimpleValue,
+    <V as Freeze>::Frozen: StarlarkValue<'static>,
 {
 }
 
-/// A trait representing Starlark values which are simple - they
-/// aren't mutable and can't contain other Starlark values.
-///
-/// Let's define a simple object, where `+x` makes the string uppercase:
-///
-/// ```
-/// use starlark::values::{Heap, StarlarkValue, Value};
-/// use starlark::{starlark_simple_value, starlark_type};
-/// use derive_more::Display;
-///
-/// #[derive(Debug, Display)]
-/// struct MyObject(String);
-/// starlark_simple_value!(MyObject);
-/// impl<'v> StarlarkValue<'v> for MyObject {
-///     starlark_type!("my_object");
-///
-///     // We can choose to implement whichever methods we want.
-///     // All other operations will result in runtime errors.
-///     fn plus(&self, heap: &'v Heap) -> anyhow::Result<Value<'v>> {
-///         Ok(heap.alloc(MyObject(self.0.to_uppercase())))
-///     }
-/// }
-/// ```
-///
-/// The [`starlark_simple_value!`] macro defines instances of [`AnyLifetime`],
-/// [`AllocValue`](crate::values::AllocValue),
-/// [`AllocFrozenValue`](crate::values::AllocFrozenValue), [`SimpleValue`] and
-/// [`UnpackValue`](crate::values::UnpackValue). It also defines a method:
-///
-/// ```
-/// # use crate::starlark::values::*;
-/// # struct MyObject;
-/// impl MyObject {
-///     pub fn from_value<'v>(x: Value<'v>) -> Option<ARef<'v, MyObject>> {
-/// # unimplemented!(
-/// # r#"
-///         ...
-/// # "#);
-///     }
-/// }
-/// ```
-///
-/// All users defining [`SimpleValue`] should use this macro.
-pub trait SimpleValue: StarlarkValue<'static> + Send + Sync {}
-
-/// Non-instantiatable type implementing [`SimpleValue`].
+/// Non-instantiatable type implementing a simple [`StarlarkValue`].
 ///
 /// Useful when implementing [`ComplexValue`], which is not meant to be frozen
 /// (e.g. when evaluating code which is never frozen or in tests).
-#[derive(Debug, AnyLifetime, Display)]
+#[derive(Debug, AnyLifetime, Display, NoSerialize)]
 #[display(fmt = "NoSimpleValue")] // Can't actually be invoked since no &self
 pub enum NoSimpleValue {}
 impl<'v> StarlarkValue<'v> for NoSimpleValue {
     starlark_type!("no_simple_value");
 }
 
-impl SimpleValue for NoSimpleValue {}
-
 /// How to put a Rust values into [`Value`]s.
 ///
-/// Every Rust value stored in a [`Value`] must implement this trait, along with
-/// either [`SimpleValue`] or [`ComplexValue`]. You _must_ implement [`ComplexValue`] if:
+/// Every Rust value stored in a [`Value`] must implement this trait.
+/// You _must_ also implement [`ComplexValue`] if:
 ///
 /// * A type is _mutable_, if you ever need to get a `&mut self` reference to it.
 /// * A type _contains_ nested Starlark [`Value`]s.
 ///
-/// Otherwise you should implement [`SimpleValue`].
-/// See those two traits for examples of how to implement them.
-///
 /// There are only two required methods of [`StarlarkValue`], namely
 /// [`get_type`](StarlarkValue::get_type)
 /// and [`get_type_value_static`](StarlarkValue::get_type_value_static).
@@ -233,11 +183,13 @@ impl SimpleValue for NoSimpleValue {}
 ///
 /// ```
 /// use starlark::values::StarlarkValue;
+/// use starlark::values::AnyLifetime;
+/// use starlark::values::NoSerialize;
 /// # use starlark::starlark_simple_value;
 /// use starlark::starlark_type;
 /// use derive_more::Display;
 ///
-/// #[derive(Debug, Display)]
+/// #[derive(Debug, Display, AnyLifetime, NoSerialize)]
 /// #[display(fmt = "Foo")]
 /// struct Foo;
 /// # starlark_simple_value!(Foo);
@@ -253,7 +205,7 @@ impl SimpleValue for NoSimpleValue {}
 /// Any additional methods that are added to this trait also need to be added to the
 /// [`StarlarkValue`] implementation in `crate::values::layout::avalue::Wrapper`. Otherwise,
 /// any implementations other than the default implementation will not be run.
-pub trait StarlarkValue<'v>: 'v + AnyLifetime<'v> + Debug + Display {
+pub trait StarlarkValue<'v>: 'v + AnyLifetime<'v> + Debug + Display + Serialize {
     /// Return a string describing the type of self, as returned by the type()
     /// function.
     ///
@@ -269,6 +221,18 @@ pub trait StarlarkValue<'v>: 'v + AnyLifetime<'v> + Debug + Display {
     where
         Self: Sized;
 
+    /// Type is special in Starlark, it is implemented differently than user defined types.
+    /// For example, some special types like `bool` cannon be heap allocated.
+    ///
+    /// This function must not be implemented outside of starlark crate.
+    #[doc(hidden)]
+    fn is_special() -> bool
+    where
+        Self: Sized,
+    {
+        false
+    }
+
     /// Is this value a match for a named type. Usually returns `true` for
     /// values matching `get_type`, but might also work for subtypes it implements.
     fn matches_type(&self, ty: &str) -> bool {
@@ -314,11 +278,6 @@ pub trait StarlarkValue<'v>: 'v + AnyLifetime<'v> + Debug + Display {
         write!(collector, "<{}...>", self.get_type()).unwrap()
     }
 
-    /// Convert the type to a JSON string.
-    fn collect_json(&self, _collector: &mut String) -> anyhow::Result<()> {
-        ValueError::unsupported(self, "collect_json()")
-    }
-
     /// Convert self to a boolean, as returned by the bool() function.
     /// The default implementation returns [`true`].
     fn to_bool(&self) -> bool {
@@ -361,6 +320,11 @@ pub trait StarlarkValue<'v>: 'v + AnyLifetime<'v> + Debug + Display {
 
     /// Compare `self` with `other` for equality.
     /// Should only return an error on excessive recursion.
+    ///
+    /// This function can only be called when it is known that self pointer
+    /// is not equal to the other pointer. Thus, an implementation may assume
+    /// that the pointers are not equal. Implementation of `equals` for some
+    /// builtin types and default implementation rely on this assumption.
     fn equals(&self, _other: Value<'v>) -> anyhow::Result<bool> {
         // Type is only equal via a pointer
         Ok(false)
@@ -375,13 +339,10 @@ pub trait StarlarkValue<'v>: 'v + AnyLifetime<'v> + Debug + Display {
 
     /// Directly invoke a function.
     /// The number of `named` and `names` arguments are guaranteed to be equal.
-    /// A direct implementation is responsible for calling [`Evaluator::with_call_stack`] to ensure
-    /// the call stack is properly updated.
     fn invoke(
         &self,
         _me: Value<'v>,
-        _location: Option<Span>,
-        _args: Arguments<'v, '_>,
+        _args: &Arguments<'v, '_>,
         _eval: &mut Evaluator<'v, '_>,
     ) -> anyhow::Result<Value<'v>> {
         ValueError::unsupported(self, "call()")
@@ -396,8 +357,7 @@ pub trait StarlarkValue<'v>: 'v + AnyLifetime<'v> + Debug + Display {
         &self,
         _me: Value<'v>,
         _this: Value<'v>,
-        _location: Option<Span>,
-        _args: Arguments<'v, '_>,
+        _args: &Arguments<'v, '_>,
         _eval: &mut Evaluator<'v, '_>,
     ) -> anyhow::Result<Value<'v>> {
         unreachable!("invoke_method should only be invoked for method or attribute");
@@ -666,7 +626,18 @@ pub trait StarlarkValue<'v>: 'v + AnyLifetime<'v> + Debug + Display {
     }
 
     /// Bitwise `|` operator.
-    fn bit_or(&self, other: Value<'v>) -> anyhow::Result<Value<'v>> {
+    ///
+    /// # Examples
+    ///
+    /// ```rust
+    /// # starlark::assert::all_true(r#"
+    /// 0xb00f | 0x0ee0 == 0xbeef
+    /// 4 | 7 == 7
+    /// {1: 2} | {3: 4} == {1: 2, 3: 4}
+    /// {1: 2} | {1: 3} == {1: 3}
+    /// # "#);
+    /// ```
+    fn bit_or(&self, other: Value<'v>, _heap: &'v Heap) -> anyhow::Result<Value<'v>> {
         ValueError::unsupported_with(self, "|", other)
     }
 
@@ -712,9 +683,11 @@ pub trait StarlarkValue<'v>: 'v + AnyLifetime<'v> + Debug + Display {
     }
 }
 
+erased_serde::serialize_trait_object!(<'v> StarlarkValue<'v>);
+
 /// Trait implemented by a value stored in arena which delegates
 /// it's operations to contained [`StarlarkValue`].
-pub(crate) trait StarlarkValueDyn<'v>: 'v {
+pub(crate) trait StarlarkValueDyn<'v>: 'v + Serialize {
     // `AValue` is not a `StarlarkValue`, but a `Wrapper` type (or `BlackHole`).
     // `static_type_xxx_of_value` operations return `TypeId` of that `StarlarkValue`,
     // which is not the same of `TypeId` of `AValue` (because `AValue` is a wrapper).
@@ -738,7 +711,6 @@ pub(crate) trait StarlarkValueDyn<'v>: 'v {
     fn documentation(&self) -> Option<DocItem>;
     fn collect_repr(&self, _collector: &mut String);
     fn collect_repr_cycle(&self, _collector: &mut String);
-    fn collect_json(&self, _collector: &mut String) -> anyhow::Result<()>;
     fn to_bool(&self) -> bool;
     fn to_int(&self) -> anyhow::Result<i32>;
     fn write_hash(&self, hasher: &mut StarlarkHasher) -> anyhow::Result<()>;
@@ -748,16 +720,14 @@ pub(crate) trait StarlarkValueDyn<'v>: 'v {
     fn invoke(
         &self,
         _me: Value<'v>,
-        _location: Option<Span>,
-        _args: Arguments<'v, '_>,
+        _args: &Arguments<'v, '_>,
         _eval: &mut Evaluator<'v, '_>,
     ) -> anyhow::Result<Value<'v>>;
     fn invoke_method(
         &self,
         me: Value<'v>,
         this: Value<'v>,
-        location: Option<Span>,
-        args: Arguments<'v, '_>,
+        args: &Arguments<'v, '_>,
         eval: &mut Evaluator<'v, '_>,
     ) -> anyhow::Result<Value<'v>>;
     fn at(&self, _index: Value<'v>, _heap: &'v Heap) -> anyhow::Result<Value<'v>>;
@@ -794,7 +764,7 @@ pub(crate) trait StarlarkValueDyn<'v>: 'v {
     fn div(&self, _other: Value<'v>, _heap: &'v Heap) -> anyhow::Result<Value<'v>>;
     fn floor_div(&self, _other: Value<'v>, _heap: &'v Heap) -> anyhow::Result<Value<'v>>;
     fn bit_and(&self, _other: Value<'v>) -> anyhow::Result<Value<'v>>;
-    fn bit_or(&self, _other: Value<'v>) -> anyhow::Result<Value<'v>>;
+    fn bit_or(&self, _other: Value<'v>, _heap: &'v Heap) -> anyhow::Result<Value<'v>>;
     fn bit_xor(&self, _other: Value<'v>) -> anyhow::Result<Value<'v>>;
     fn left_shift(&self, _other: Value<'v>) -> anyhow::Result<Value<'v>>;
     fn right_shift(&self, _other: Value<'v>) -> anyhow::Result<Value<'v>>;
@@ -802,3 +772,5 @@ pub(crate) trait StarlarkValueDyn<'v>: 'v {
     fn set_at(&self, _index: Value<'v>, _new_value: Value<'v>) -> anyhow::Result<()>;
     fn set_attr(&self, _attribute: &str, _new_value: Value<'v>) -> anyhow::Result<()>;
 }
+
+erased_serde::serialize_trait_object!(<'v> StarlarkValueDyn<'v>);
