@@ -18,13 +18,14 @@
 //! The dictionary type, a mutable associative-map, which iterates in insertion order.
 
 use std::{
+    any::TypeId,
     cell::{Ref, RefCell, RefMut},
     fmt,
     fmt::{Debug, Display},
     hash::{Hash, Hasher},
     intrinsics::unlikely,
     marker::PhantomData,
-    ops::Deref,
+    ops::{Deref, DerefMut},
 };
 
 use gazebo::{
@@ -33,6 +34,7 @@ use gazebo::{
     coerce::{coerce, coerce_ref, Coerce},
 };
 use indexmap::Equivalent;
+use serde::{ser::SerializeMap, Serialize};
 
 use crate::{
     self as starlark,
@@ -41,12 +43,12 @@ use crate::{
     values::{
         comparison::equals_small_map, display::display_keyed_container, error::ValueError,
         iter::ARefIterator, string::hash_string_value, AllocFrozenValue, AllocValue, Freeze,
-        Freezer, FrozenHeap, FrozenStringValue, FrozenValue, Heap, SimpleValue, StarlarkValue,
-        StringValue, Trace, UnpackValue, Value, ValueLike,
+        Freezer, FrozenHeap, FrozenStringValue, FrozenValue, Heap, StarlarkValue, StringValue,
+        Trace, UnpackValue, Value, ValueLike,
     },
 };
 
-#[derive(Clone, Default, Trace, Debug)]
+#[derive(Clone, Default, Trace, Debug, AnyLifetime)]
 struct DictGen<T>(T);
 
 impl<'v, T: DictLike<'v>> Display for DictGen<T> {
@@ -55,8 +57,14 @@ impl<'v, T: DictLike<'v>> Display for DictGen<T> {
     }
 }
 
+impl<'v> Display for Dict<'v> {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        display_keyed_container(f, "{", "}", ": ", self.iter())
+    }
+}
+
 /// Define the list type. See [`Dict`] and [`FrozenDict`] as the two possible representations.
-#[derive(Clone, Default, Trace, Debug)]
+#[derive(Clone, Default, Trace, Debug, AnyLifetime)]
 #[repr(transparent)]
 pub struct Dict<'v> {
     /// The data stored by the dictionary. The keys must all be hashable values.
@@ -71,11 +79,6 @@ pub struct FrozenDict {
     content: SmallMap<FrozenValue, FrozenValue>,
 }
 
-unsafe impl<'v> AnyLifetime<'v> for DictGen<RefCell<Dict<'v>>> {
-    any_lifetime_body!(DictGen<RefCell<Dict<'static>>>);
-}
-any_lifetime!(DictGen<FrozenDict>);
-
 unsafe impl<'v> Coerce<Dict<'v>> for FrozenDict {}
 
 impl<'v> AllocValue<'v> for Dict<'v> {
@@ -90,20 +93,55 @@ impl AllocFrozenValue for FrozenDict {
     }
 }
 
-impl SimpleValue for DictGen<FrozenDict> {}
+/// Borrowed `Dict`.
+pub struct DictRef<'v> {
+    aref: ARef<'v, Dict<'v>>,
+}
+
+/// Mutably borrowed `Dict`.
+pub struct DictMut<'v> {
+    aref: RefMut<'v, Dict<'v>>,
+}
+
+impl<'v> Deref for DictRef<'v> {
+    type Target = Dict<'v>;
+
+    fn deref(&self) -> &Self::Target {
+        &self.aref
+    }
+}
+
+impl<'v> Deref for DictMut<'v> {
+    type Target = Dict<'v>;
+
+    fn deref(&self) -> &Self::Target {
+        &self.aref
+    }
+}
+
+impl<'v> DerefMut for DictMut<'v> {
+    fn deref_mut(&mut self) -> &mut Self::Target {
+        &mut self.aref
+    }
+}
 
 impl<'v> Dict<'v> {
-    pub fn from_value(x: Value<'v>) -> Option<ARef<'v, Self>> {
+    /// Downcast the value to a dict.
+    pub fn from_value(x: Value<'v>) -> Option<DictRef<'v>> {
         if x.unpack_frozen().is_some() {
-            x.downcast_ref::<DictGen<FrozenDict>>()
-                .map(|x| ARef::new_ptr(coerce_ref(&x.0)))
+            x.downcast_ref::<DictGen<FrozenDict>>().map(|x| DictRef {
+                aref: ARef::new_ptr(coerce_ref(&x.0)),
+            })
         } else {
             let ptr = x.downcast_ref::<DictGen<RefCell<Dict<'v>>>>()?;
-            Some(ARef::new_ref(ptr.0.borrow()))
+            Some(DictRef {
+                aref: ARef::new_ref(ptr.0.borrow()),
+            })
         }
     }
 
-    pub fn from_value_mut(x: Value<'v>) -> anyhow::Result<Option<RefMut<'v, Self>>> {
+    /// Downcast the value to a mutable dict reference.
+    pub fn from_value_mut(x: Value<'v>) -> anyhow::Result<Option<DictMut>> {
         if unlikely(x.unpack_frozen().is_some()) {
             return Err(ValueError::CannotMutateImmutableValue.into());
         }
@@ -111,19 +149,29 @@ impl<'v> Dict<'v> {
         match ptr {
             None => Ok(None),
             Some(ptr) => match ptr.0.try_borrow_mut() {
-                Ok(x) => Ok(Some(x)),
+                Ok(x) => Ok(Some(DictMut { aref: x })),
                 Err(_) => Err(ValueError::MutationDuringIteration.into()),
             },
         }
     }
+
+    pub(crate) fn is_dict_type(x: TypeId) -> bool {
+        x == TypeId::of::<DictGen<FrozenDict>>()
+            || x == TypeId::of::<DictGen<RefCell<Dict<'static>>>>()
+    }
+
+    pub(crate) unsafe fn from_value_unchecked_mut(x: Value<'v>) -> RefMut<'v, Self> {
+        let dict = &x.downcast_ref_unchecked::<DictGen<RefCell<Dict<'v>>>>().0;
+        dict.borrow_mut()
+    }
 }
 
-impl<'v> UnpackValue<'v> for ARef<'v, Dict<'v>> {
+impl<'v> UnpackValue<'v> for DictRef<'v> {
     fn expected() -> String {
         "dict".to_owned()
     }
 
-    fn unpack_value(value: Value<'v>) -> Option<ARef<'v, Dict<'v>>> {
+    fn unpack_value(value: Value<'v>) -> Option<DictRef<'v>> {
         Dict::from_value(value)
     }
 }
@@ -163,6 +211,7 @@ impl<'v> Dict<'v> {
     /// The result of calling `type()` on dictionaries.
     pub const TYPE: &'static str = "dict";
 
+    /// Dict type string as Starlark frozen string value.
     pub fn get_type_value_static() -> FrozenStringValue {
         DictGen::<FrozenDict>::get_type_value_static()
     }
@@ -172,16 +221,18 @@ impl<'v> Dict<'v> {
         Self { content }
     }
 
+    /// Number of elements in the dict.
     pub fn len(&self) -> usize {
         self.content.len()
     }
 
+    /// Is the dict empty?
     pub fn is_empty(&self) -> bool {
         self.content.is_empty()
     }
 
     /// Iterate through the key/value pairs in the dictionary.
-    pub fn iter<'a>(&'a self) -> impl Iterator<Item = (Value<'v>, Value<'v>)> + 'a {
+    pub fn iter<'a>(&'a self) -> impl ExactSizeIterator<Item = (Value<'v>, Value<'v>)> + 'a {
         self.content.iter().map(|(l, r)| (*l, *r))
     }
 
@@ -211,6 +262,7 @@ impl<'v> Dict<'v> {
         Ok(self.get_hashed(key.get_hashed()?))
     }
 
+    /// Lookup the value by the given prehashed key.
     pub fn get_hashed(&self, key: Hashed<Value<'v>>) -> Option<Value<'v>> {
         self.content.get_hashed(key.borrow()).copied()
     }
@@ -252,18 +304,22 @@ impl<'v> Dict<'v> {
         })
     }
 
+    /// Reserve capacity to insert `additional` elements without reallocating.
     pub fn reserve(&mut self, additional: usize) {
         self.content.reserve(additional);
     }
 
+    /// Insert a key/value pair into the dictionary.
     pub fn insert_hashed(&mut self, key: Hashed<Value<'v>>, value: Value<'v>) {
         self.content.insert_hashed(key, value);
     }
 
+    /// Remove given key from the dictionary.
     pub fn remove_hashed(&mut self, key: Hashed<Value<'v>>) -> Option<Value<'v>> {
         self.content.remove_hashed(key.borrow())
     }
 
+    /// Remove all elements from the dictionary.
     pub fn clear(&mut self) {
         self.content.clear();
     }
@@ -346,6 +402,11 @@ impl<'v> DictLike<'v> for FrozenDict {
     }
 }
 
+pub(crate) fn dict_methods() -> Option<&'static Methods> {
+    static RES: MethodsStatic = MethodsStatic::new();
+    RES.methods(crate::stdlib::dict::dict_methods)
+}
+
 impl<'v, T: DictLike<'v>> StarlarkValue<'v> for DictGen<T>
 where
     Self: AnyLifetime<'v>,
@@ -353,8 +414,7 @@ where
     starlark_type!(Dict::TYPE);
 
     fn get_methods(&self) -> Option<&'static Methods> {
-        static RES: MethodsStatic = MethodsStatic::new();
-        RES.methods(crate::stdlib::dict::dict_methods)
+        dict_methods()
     }
 
     fn collect_repr(&self, r: &mut String) {
@@ -375,20 +435,6 @@ where
         collector.push_str("{...}");
     }
 
-    fn collect_json(&self, collector: &mut String) -> anyhow::Result<()> {
-        collector.push('{');
-        for (i, (k, v)) in self.0.content().iter().enumerate() {
-            if i != 0 {
-                collector.push_str(", ");
-            }
-            k.collect_json(collector)?;
-            collector.push_str(": ");
-            v.collect_json(collector)?;
-        }
-        collector.push('}');
-        Ok(())
-    }
-
     fn to_bool(&self) -> bool {
         !self.0.content().is_empty()
     }
@@ -448,6 +494,37 @@ where
         let index = index.get_hashed()?;
         self.0.set_at(index, alloc_value)
     }
+
+    fn bit_or(&self, rhs: Value<'v>, heap: &'v Heap) -> anyhow::Result<Value<'v>> {
+        let rhs = Dict::from_value(rhs)
+            .map_or_else(|| ValueError::unsupported_with(self, "|", rhs), Ok)?;
+        if self.0.content().is_empty() {
+            return Ok(heap.alloc(rhs.clone()));
+        }
+        // Might be faster if we preallocate the capacity, but then copying in the LHS
+        // is more expensive and might oversize given the behaviour on duplicates.
+        // If this becomes a bottleneck, benchmark.
+        let mut items = self.0.content().clone();
+        for (k, v) in rhs.iter_hashed() {
+            items.insert_hashed(k, v);
+        }
+        Ok(heap.alloc(Dict::new(items)))
+    }
+}
+
+impl<'v, T: DictLike<'v>> Serialize for DictGen<T> {
+    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
+    where
+        S: serde::Serializer,
+    {
+        let mut map_serializer = serializer.serialize_map(Some(self.0.content().len()))?;
+
+        for (k, v) in self.0.content().iter() {
+            map_serializer.serialize_entry(k, v)?;
+        }
+
+        map_serializer.end()
+    }
 }
 
 impl<'v, K: UnpackValue<'v> + Hash + Eq, V: UnpackValue<'v>> UnpackValue<'v> for SmallMap<K, V> {
@@ -473,6 +550,7 @@ pub struct DictOf<'v, K: UnpackValue<'v>, V: UnpackValue<'v>> {
 }
 
 impl<'v, K: UnpackValue<'v>, V: UnpackValue<'v>> DictOf<'v, K, V> {
+    /// Get all the elements.
     // This should return an iterator, but it is not trivial to do with `ARef`.
     pub fn collect_entries(&self) -> Vec<(K, V)> {
         Dict::from_value(self.value)
@@ -489,6 +567,7 @@ impl<'v, K: UnpackValue<'v>, V: UnpackValue<'v>> DictOf<'v, K, V> {
 }
 
 impl<'v, K: UnpackValue<'v> + Hash + Eq, V: UnpackValue<'v>> DictOf<'v, K, V> {
+    /// Collect all the elements to a fresh `SmallMap`.
     pub fn to_dict(&self) -> SmallMap<K, V> {
         Dict::from_value(self.value)
             .expect("already validated as a dict")
@@ -555,12 +634,12 @@ b1 and b2 and b3
     #[test]
     fn test_get_str() -> anyhow::Result<()> {
         let heap = Heap::new();
-        let k1 = heap.alloc_str_hashed("hello");
-        let k2 = heap.alloc_str_hashed("world");
+        let k1 = heap.alloc_str("hello").get_hashed();
+        let k2 = heap.alloc_str("world").get_hashed();
         let mut sm = SmallMap::new();
         sm.insert_hashed(k1, Value::new_int(12));
         sm.insert_hashed(k2, Value::new_int(56));
-        let d = Dict::new(sm);
+        let d = Dict::new(coerce(sm));
 
         assert_eq!(d.get(heap.alloc("hello"))?.unwrap().unpack_int(), Some(12));
         assert_eq!(d.get(heap.alloc("foo"))?, None);
