@@ -34,6 +34,7 @@ use gazebo::{
     coerce::{coerce, coerce_ref, Coerce},
     prelude::*,
 };
+use serde::{ser::SerializeSeq, Serialize};
 
 use crate::{
     self as starlark,
@@ -49,7 +50,7 @@ use crate::{
     },
 };
 
-#[derive(Clone, Default, Trace, Debug)]
+#[derive(Clone, Default, Trace, Debug, AnyLifetime)]
 #[repr(transparent)]
 pub(crate) struct ListGen<T>(pub(crate) T);
 
@@ -78,11 +79,12 @@ impl<'v> Debug for FrozenList {
 }
 
 impl ListGen<FrozenList> {
-    pub fn offset_of_content() -> usize {
+    pub(crate) fn offset_of_content() -> usize {
         memoffset::offset_of!(FrozenList, content)
     }
 }
 
+/// Reference to list content.
 #[repr(transparent)]
 #[derive(Coerce)]
 pub struct ListRef<'v> {
@@ -94,6 +96,7 @@ impl<'v> ListRef<'v> {
         coerce(slice)
     }
 
+    /// List elements.
     pub fn content(&self) -> &[Value<'v>] {
         &self.content
     }
@@ -107,12 +110,8 @@ impl<'v> ListRef<'v> {
     }
 }
 
-unsafe impl<'v> AnyLifetime<'v> for ListGen<List<'v>> {
-    any_lifetime_body!(ListGen<List<'static>>);
-}
-any_lifetime!(ListGen<FrozenList>);
-
 impl<'v> List<'v> {
+    /// Downcast the list.
     pub fn from_value(x: Value<'v>) -> Option<&'v ListRef<'v>> {
         if x.unpack_frozen().is_some() {
             x.downcast_ref::<ListGen<FrozenList>>()
@@ -137,6 +136,12 @@ impl<'v> List<'v> {
         }
     }
 
+    pub(crate) unsafe fn from_value_unchecked_mut(x: Value<'v>) -> &'v Self {
+        let list = x.downcast_ref_unchecked::<ListGen<List<'v>>>();
+        debug_assert!(list.0.check_can_mutate().is_ok());
+        &list.0
+    }
+
     pub(crate) fn is_list_type(x: TypeId) -> bool {
         x == TypeId::of::<ListGen<List>>() || x == TypeId::of::<ListGen<FrozenList>>()
     }
@@ -299,6 +304,7 @@ impl<'v> List<'v> {
     /// The result of calling `type()` on lists.
     pub const TYPE: &'static str = "list";
 
+    /// Type of list as frozen string value.
     pub fn get_type_value_static() -> FrozenStringValue {
         ListGen::<FrozenList>::get_type_value_static()
     }
@@ -432,15 +438,26 @@ pub(crate) fn display_list(xs: &[Value], f: &mut fmt::Formatter<'_>) -> fmt::Res
     display_container(f, "[", "]", xs.iter())
 }
 
+pub(crate) fn list_methods() -> Option<&'static Methods> {
+    static RES: MethodsStatic = MethodsStatic::new();
+    RES.methods(crate::stdlib::list::list_methods)
+}
+
 impl<'v, T: ListLike<'v>> StarlarkValue<'v> for ListGen<T>
 where
     Self: AnyLifetime<'v> + Display,
 {
     starlark_type!(List::TYPE);
 
+    fn is_special() -> bool
+    where
+        Self: Sized,
+    {
+        true
+    }
+
     fn get_methods(&self) -> Option<&'static Methods> {
-        static RES: MethodsStatic = MethodsStatic::new();
-        RES.methods(crate::stdlib::list::list_methods)
+        list_methods()
     }
 
     fn collect_repr(&self, s: &mut String) {
@@ -459,18 +476,6 @@ where
         collector.push_str("[...]");
     }
 
-    fn collect_json(&self, collector: &mut String) -> anyhow::Result<()> {
-        collector.push('[');
-        for (i, e) in self.0.content().iter().enumerate() {
-            if i != 0 {
-                collector.push(',');
-            }
-            e.collect_json(collector)?;
-        }
-        collector.push(']');
-        Ok(())
-    }
-
     fn to_bool(&self) -> bool {
         !self.0.content().is_empty()
     }
@@ -564,6 +569,21 @@ where
     }
 }
 
+impl<'v, T: ListLike<'v>> Serialize for ListGen<T> {
+    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
+    where
+        S: serde::Serializer,
+    {
+        let mut seq_serializer = serializer.serialize_seq(Some(self.0.content().len()))?;
+
+        for e in self.0.content().iter() {
+            seq_serializer.serialize_element(&e.to_value())?;
+        }
+
+        seq_serializer.end()
+    }
+}
+
 /// Like `ValueOf`, but only validates item types; does not construct or store a
 /// vec. Use `to_vec` to get a Vec.
 pub struct ListOf<'v, V: UnpackValue<'v>> {
@@ -572,6 +592,7 @@ pub struct ListOf<'v, V: UnpackValue<'v>> {
 }
 
 impl<'v, V: UnpackValue<'v>> ListOf<'v, V> {
+    /// Collect the list elements into a `Vec`.
     pub fn to_vec(&self) -> Vec<V> {
         List::from_value(self.value)
             .expect("already validated as a list")
