@@ -29,17 +29,15 @@ use std::{
 };
 
 use gazebo::{any::AnyLifetime, prelude::*};
+use serde::Serialize;
 
 use crate::{
-    collections::{BorrowHashed, SmallHashResult, StarlarkHasher},
+    collections::{BorrowHashed, StarlarkHashValue, StarlarkHasher},
     environment::{Methods, MethodsStatic},
     values::{
         index::apply_slice,
         string::repr::string_repr,
-        types::{
-            none::NoneOr,
-            string::{fast_string::StrIndices, json::json_escape},
-        },
+        types::{none::NoneOr, string::fast_string::StrIndices},
         Heap, StarlarkValue, UnpackValue, Value, ValueError,
     },
 };
@@ -48,7 +46,6 @@ mod alloc_unpack;
 pub(crate) mod fast_string;
 pub(crate) mod interpolation;
 pub(crate) mod iter;
-mod json;
 mod repr;
 pub(crate) mod simd;
 
@@ -77,6 +74,7 @@ impl Add for CharIndex {
 pub const STRING_TYPE: &str = "string";
 
 #[repr(C)] // We want the body to come after len
+#[derive(AnyLifetime)]
 pub(crate) struct StarlarkStrN<const N: usize> {
     // Lazily-initialized cached hash code.
     pub(crate) hash: atomic::AtomicU32,
@@ -88,10 +86,6 @@ pub(crate) struct StarlarkStrN<const N: usize> {
     pub(crate) body: [u8; N],
 }
 
-unsafe impl<'a, const N: usize> AnyLifetime<'a> for StarlarkStrN<N> {
-    any_lifetime_body!(StarlarkStrN<N>);
-}
-
 /// A pointer to this type represents a Starlark string.
 /// Use of this type is discouraged and not considered stable.
 #[derive(AnyLifetime)]
@@ -127,6 +121,7 @@ impl StarlarkStr {
         }
     }
 
+    /// Get a Rust string refence from this Starlark string.
     pub fn unpack(&self) -> &str {
         unsafe {
             let slice = slice::from_raw_parts(self.str.body.as_ptr(), self.str.len as usize);
@@ -134,11 +129,12 @@ impl StarlarkStr {
         }
     }
 
-    pub(crate) fn get_small_hash_result(&self) -> SmallHashResult {
+    /// Get cached hash value or compute if it is not cached yet.
+    pub fn get_hash(&self) -> StarlarkHashValue {
         // Note relaxed load and store are practically non-locking memory operations.
         let hash = self.str.hash.load(atomic::Ordering::Relaxed);
         if hash != 0 {
-            SmallHashResult::new_unchecked(hash)
+            StarlarkHashValue::new_unchecked(hash)
         } else {
             let mut s = StarlarkHasher::new();
             hash_string_value(self.unpack(), &mut s);
@@ -149,33 +145,32 @@ impl StarlarkStr {
         }
     }
 
+    /// Rust string reference along with its hash value.
     pub fn as_str_hashed(&self) -> BorrowHashed<str> {
-        BorrowHashed::new_unchecked(self.get_small_hash_result(), self.unpack())
+        BorrowHashed::new_unchecked(self.get_hash(), self.unpack())
     }
 
+    /// String length, in bytes.
     pub fn len(&self) -> usize {
         self.str.len as usize
     }
 
+    /// Is this string empty?
     pub fn is_empty(&self) -> bool {
         self.str.len == 0
     }
 
-    pub fn offset_of_content() -> usize {
+    pub(crate) fn offset_of_content() -> usize {
         memoffset::offset_of!(StarlarkStrN<0>, body)
     }
 }
 
 /// How to hash a string in a way that is compatible with Value
+#[inline]
 pub(crate) fn hash_string_value<H: Hasher>(x: &str, state: &mut H) {
     x.hash(state)
 }
 
-/// Hash a string in a way compatible with Value
-pub(crate) fn hash_string_result(x: &str) -> SmallHashResult {
-    SmallHashResult::new(x)
-}
-
 impl Display for StarlarkStr {
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         // We could either accumulate straight into the buffer (can't preallocate, virtual call on each character)
@@ -187,14 +182,18 @@ impl Display for StarlarkStr {
     }
 }
 
+pub(crate) fn str_methods() -> Option<&'static Methods> {
+    static RES: MethodsStatic = MethodsStatic::new();
+    RES.methods(crate::stdlib::string::string_methods)
+}
+
 // We don't actually use `str` on the heap, but it's helpful to have an instance
 // for people who want to get access to repr/json "like string"
 impl<'v> StarlarkValue<'v> for str {
     starlark_type!(STRING_TYPE);
 
     fn get_methods(&self) -> Option<&'static Methods> {
-        static RES: MethodsStatic = MethodsStatic::new();
-        RES.methods(crate::stdlib::string::string_methods)
+        str_methods()
     }
 
     fn collect_repr(&self, buffer: &mut String) {
@@ -202,11 +201,6 @@ impl<'v> StarlarkValue<'v> for str {
         string_repr(self, buffer)
     }
 
-    fn collect_json(&self, collector: &mut String) -> anyhow::Result<()> {
-        json_escape(self, collector);
-        Ok(())
-    }
-
     fn to_bool(&self) -> bool {
         !self.is_empty()
     }
@@ -264,11 +258,7 @@ impl<'v> StarlarkValue<'v> for str {
 
     fn is_in(&self, other: Value) -> anyhow::Result<bool> {
         let s = <&str>::unpack_param(other)?;
-        if s.len() == 1 {
-            Ok(self.as_bytes().contains(&s.as_bytes()[0]))
-        } else {
-            Ok(self.contains(s))
-        }
+        Ok(fast_string::contains(self, s))
     }
 
     fn slice(
@@ -297,8 +287,8 @@ impl<'v> StarlarkValue<'v> for str {
         let (start, stop) = (start_stop_to_none_or(start)?, start_stop_to_none_or(stop)?);
 
         match fast_string::convert_str_indices(self, start, stop) {
-            Some(StrIndices { haystack, .. }) => Ok(heap.alloc_str(haystack)),
-            None => Ok(heap.alloc_str("")),
+            Some(StrIndices { haystack, .. }) => Ok(heap.alloc_str(haystack).to_value()),
+            None => Ok(heap.alloc_str("").to_value()),
         }
     }
 
@@ -307,7 +297,7 @@ impl<'v> StarlarkValue<'v> for str {
             if self.is_empty() {
                 Ok(other)
             } else {
-                Ok(heap.alloc_str_concat(self, other_str))
+                Ok(heap.alloc_str_concat(self, other_str).to_value())
             }
         } else {
             ValueError::unsupported_with(self, "+", other)
@@ -331,9 +321,16 @@ impl<'v> StarlarkValue<'v> for str {
 impl<'v> StarlarkValue<'v> for StarlarkStr {
     starlark_type!(STRING_TYPE);
 
+    fn is_special() -> bool
+    where
+        Self: Sized,
+    {
+        true
+    }
+
     fn write_hash(&self, hasher: &mut StarlarkHasher) -> anyhow::Result<()> {
         // Don't defer to str because we cache the Hash in StarlarkStr
-        hasher.write_u32(self.get_small_hash_result().get());
+        hasher.write_u32(self.get_hash().get());
         Ok(())
     }
 
@@ -352,10 +349,6 @@ impl<'v> StarlarkValue<'v> for StarlarkStr {
         self.unpack().collect_repr(collector)
     }
 
-    fn collect_json(&self, collector: &mut String) -> anyhow::Result<()> {
-        self.unpack().collect_json(collector)
-    }
-
     fn to_bool(&self) -> bool {
         self.unpack().to_bool()
     }
@@ -403,6 +396,15 @@ impl<'v> StarlarkValue<'v> for StarlarkStr {
     }
 }
 
+impl Serialize for StarlarkStr {
+    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
+    where
+        S: serde::Serializer,
+    {
+        serializer.serialize_str(self.unpack())
+    }
+}
+
 #[cfg(test)]
 mod tests {
     use crate::{
@@ -453,7 +455,7 @@ mod tests {
         let heap = Heap::new();
         for x in EXAMPLES {
             assert_eq!(
-                heap.alloc_str_hashed(x).hash(),
+                heap.alloc_str(x).get_hashed().hash(),
                 heap.alloc(*x).get_hashed().unwrap().hash()
             );
         }
@@ -489,7 +491,7 @@ len("😿") == 1
     fn test_slice_string() {
         let heap = Heap::new();
         for example in EXAMPLES {
-            let s = heap.alloc_str(example);
+            let s = heap.alloc_str(example).to_value();
             for i in -5..=6 {
                 for j in -5..=6 {
                     let start = if i == 6 {
